/* main.c -- Main routine for WTG model.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <quickcandis.h>
#include "diabat3.h"
#include "wallclock.h"
#define VERSION "column-112"
#define PSURF 100000.
#define RANHEIGHTSCALE 3000.

FILE *startfile;

FILE *outfile;


void getprofile(double time, double *th0, double *rt0,
		double *vx0, double *vy0);
void readprofile(long nz1, char *profname);

 
/* doublebuff -- This routine uses malloc to allocate double buffer
 * space.  The argument nelem defines the number of doubleing
 * point elements.  It returns a double pointer to the space.  If malloc
 * fails, getbuff dies with an error message.
 */
double *doublebuff(long nelem)
{
  double *point;
  if ((point = (double*)malloc((unsigned)(sizeof(double)*nelem)))
       == (double*)NULL) {
    fprintf(stderr,"doublebuff: can't get %d words\n",nelem);
    exit(1);
  }
  return(point);
}

/* fbuf -- Free a buffer obtained with doublebuff */
void fbuf(double *ptr)
{
  free((char*)ptr);
}





/* routines for decoding input lines */

/* getfvar -- Get a double variable from the start file */
static double getfvar(char *vname)
{
  char buffer[121], name[121], cvalue[121];

  if (fgets(buffer, 121, startfile) == NULL) {
    fprintf(stderr,"getfvar: unexpected end of file on %s\n",vname);
    exit(1);
  }
  if (sscanf(buffer,"%s %s",name,cvalue) != 2) {
    fprintf(stderr,"getfvar: format error for parameter %s\n",vname);
    exit(1);
  }
  if (strcmp(vname,name) != 0) {
    fprintf(stderr,"getfvar: expecting parameter %s, got %s\n",vname,name);
    exit(1);
  }
  return atof(cvalue);
}

/* getivar -- Get an integer variable from the start file */
static int getivar(char *vname)
{
  char buffer[121], name[121], cvalue[121];

  if (fgets(buffer, 121, startfile) == NULL) {
    fprintf(stderr,"getivar: unexpected end of file on %s\n",vname);
    exit(1);
  }
  if (sscanf(buffer,"%s %s",name,cvalue) != 2) {
    fprintf(stderr,"getivar: format error for parameter %s\n",vname);
    exit(1);
  }
  if (strcmp(vname,name) != 0) {
    fprintf(stderr,"getivar: expecting %s, got %s\n",vname,name);
    exit(1);
  }
  return atoi(cvalue);
}

/* getsvar -- Get a string variable from the start file */
static void getsvar(char *vname,char *cvalue)
{
  char buffer[121], name[121];

  if (fgets(buffer, 121, startfile) == NULL) {
    fprintf(stderr,"getsvar: unexpected end of file on %s\n",vname);
    exit(1);
  }
  if (sscanf(buffer,"%s %s",name,cvalue) != 2) {
    fprintf(stderr,"getsvar: format error for parameter %s\n",vname);
    exit(1);
  }
}





/* fillfield -- Get the data from a specified field */
static void fillfield(int candishandle, long nz, int nfields, char *fname,
	       double *fvals)
{
  int i, j;
  float *fpointer;

  /* see if there is a field by the desired name */
  for (i = 0; i < nfields; i++) {
    if (strcmp(candisfnameget(candishandle, i), fname) == 0) {
      /* does it have the right size? */
      if (candisfsizeget(candishandle, i) != nz) {
	fprintf(stderr,"conv: initial profilefield %s has wrong size\n",
		fname);
	exit(1);
      }

      /* transfer the values */
      fpointer = candisfvalsget(candishandle, i);
      for (j = 0; j < nz; j++) fvals[j] = fpointer[j];

      /* we're done */
      return;
    }
  }

  /* nope -- fail */
  fprintf(stderr,"conv: no initial profile field named %s\n",fname);
  exit(1);
}

/* docandis -- Make a Candis file */
static void docandis(char *basename, char *profname, char *startfilename,
		     int it, int nz, double dz, double dt,
		     double time, double *dzarray, double *zcarray,
		     double *rho, double *th, double *the, double *thes,
		     double *rt, double *rsat, double *p, double *vx,
		     double *vy, double *th0, double *the0, double *thes0,
		     double *rho0, double *rt0, double *vx0, double *vy0,
		     double *sthe, double *srt, double *sth, double *sthe_cu,
		     double *sth_par, double *srt_cu, double *irflux,
		     double *svx_cu, double *svy_cu, double *pi,
		     double *vzwtg, double *dvzwtgdz, double *adjmask,
		     double *svx, double *svy, double *srad, double *theflux,
		     double *rtflux, double *vxflux, double *vyflux,
		     double *rain, double *srain,
		     double *erain, double *convthrot, double *fluxthrot,
		     double *cvcaug, double *ctopdeep, double *ctopshallow,
		     double *stemp, double *swater, double *basis,
		     double *dtempdt1, double *dqdt1)
{
  int handle, dhz;
  char cline[81];

  handle = candiscreate(basename, it);
  sprintf(cline,"paramtest: %s",VERSION);
  candiscomment(handle,cline);
  sprintf(cline,"  startfile = %s",startfilename);
  candiscomment(handle,cline);
  sprintf(cline,"  profname = %s",profname);
  candiscomment(handle,cline);
  dhz = dcandisidim(handle, "z", nz, zcarray);
  dcandisfld(handle, "time", 0, 0, 0, 0, 0, &time);
  dcandisfld(handle, "dzarray", 1, dhz, 0, 0, 0, dzarray);
  dcandisfld(handle, "zcarray", 1, dhz, 0, 0, 0, zcarray);
  dcandisfld(handle, "th0", 1, dhz, 0, 0, 0, th0);
  dcandisfld(handle, "the0", 1, dhz, 0, 0, 0, the0);
  dcandisfld(handle, "thes0", 1, dhz, 0, 0, 0, thes0);
  dcandisfld(handle, "rho0", 1, dhz, 0, 0, 0, rho0);
  dcandisfld(handle, "rt0", 1, dhz, 0, 0, 0, rt0);
  dcandisfld(handle, "vx0", 1, dhz, 0, 0, 0, vx0);
  dcandisfld(handle, "vy0", 1, dhz, 0, 0, 0, vy0);
  dcandisfld(handle, "sthe", 1, dhz, 0, 0, 0, sthe);
  dcandisfld(handle, "srt", 1, dhz, 0, 0, 0, srt);
  dcandisfld(handle, "sth", 1, dhz, 0, 0, 0, sth);
  dcandisfld(handle, "sthe_cu", 1, dhz, 0, 0, 0, sthe_cu);
  dcandisfld(handle, "sth_par", 1, dhz, 0, 0, 0, sth_par);
  dcandisfld(handle, "srt_cu", 1, dhz, 0, 0, 0, srt_cu);
  dcandisfld(handle, "irflux", 1, dhz, 0, 0, 0, irflux);
  dcandisfld(handle, "svx_cu", 1, dhz, 0, 0, 0, svx_cu);
  dcandisfld(handle, "svy_cu", 1, dhz, 0, 0, 0, svy_cu);
  dcandisfld(handle, "pi", 1, dhz, 0, 0, 0, pi);
  dcandisfld(handle, "vzwtg", 1, dhz, 0, 0, 0, vzwtg);
  dcandisfld(handle, "dvzwtgdz", 1, dhz, 0, 0, 0, dvzwtgdz);
  dcandisfld(handle, "adjmask", 1, dhz, 0, 0, 0, adjmask);
  dcandisfld(handle, "svx", 1, dhz, 0, 0, 0, svx);
  dcandisfld(handle, "svy", 1, dhz, 0, 0, 0, svy);
  dcandisfld(handle, "srad", 1, dhz, 0, 0, 0, srad);
  dcandisfld(handle, "theflux", 0, 0, 0, 0, 0, theflux);
  dcandisfld(handle, "rtflux", 0, 0, 0, 0, 0, rtflux);
  dcandisfld(handle, "vxflux", 0, 0, 0, 0, 0, vxflux);
  dcandisfld(handle, "vyflux", 0, 0, 0, 0, 0, vyflux);
  dcandisfld(handle, "rainflux", 0, 0, 0, 0, 0, rain);
  dcandisfld(handle, "srain", 0, 0, 0, 0, 0, srain);
  dcandisfld(handle, "erain", 0, 0, 0, 0, 0, erain);
  dcandisfld(handle, "convthrot", 0, 0, 0, 0, 0, convthrot);
  dcandisfld(handle, "fluxthrot", 0, 0, 0, 0, 0, fluxthrot);
  dcandisfld(handle, "cvcaug", 0, 0, 0, 0, 0, cvcaug);
  dcandisfld(handle, "ctopdeep", 0, 0, 0, 0, 0, ctopdeep);
  dcandisfld(handle, "ctopshallow", 0, 0, 0, 0, 0, ctopshallow);
  dcandisfld(handle, "stemp", 0, 0, 0, 0, 0, stemp);
  dcandisfld(handle, "swater", 0, 0, 0, 0, 0, swater);
  dcandisfld(handle, "rho", 1, dhz, 0, 0, 0, rho);
  dcandisfld(handle, "th", 1, dhz, 0, 0, 0, th);
  dcandisfld(handle, "the", 1, dhz, 0, 0, 0, the);
  dcandisfld(handle, "thes", 1, dhz, 0, 0, 0, thes);
  dcandisfld(handle, "rt", 1, dhz, 0, 0, 0, rt);
  dcandisfld(handle, "rsat", 1, dhz, 0, 0, 0, rsat);
  dcandisfld(handle, "p", 1, dhz, 0, 0, 0, p);
  dcandisfld(handle, "vx", 1, dhz, 0, 0, 0, vx);
  dcandisfld(handle, "vy", 1, dhz, 0, 0, 0, vy);
  dcandisfld(handle, "basis", 1, dhz, 0, 0, 0, basis);
  dcandisfld(handle, "dtempdt1", 1, dhz, 0, 0, 0, dtempdt1);
  dcandisfld(handle, "dqdt1", 1, dhz, 0, 0, 0, dqdt1);
  candiswrite(handle);
}





/* hydrostatic -- Compute initial hydrostatic profiles of density,
 * potential temperature, and mixing ratio on cell centers
 * starting with the surface potential temperature and the surface pressure.
 * There are assumptions about the lapse rate and the tropopause height.
 */
static void hydrostatic(int nz, double dz, double psurf, double tsurf, double rh,
			double tpause, double stab, double *th0, double *the0,
			double *thes0, double *rho0, double *rt0, double *pc0,
			double *zcarray)
{
  int iz;
  double *p0, *pi0;
  double z, grate, cpi0;

  /* allocate space */
  p0 = doublebuff(nz + 1);
  pi0 = doublebuff(nz + 1);

  /* integrate the hydrostatic equation to get pressure on cell edges
   * and density and potential temperature on cell centers */
  th0[0] = tsurf;
  pi0[0] = pifromp(psurf);
  p0[0] = pfrompi(pi0[0]);
  for (iz = 1; iz <= nz; iz++) {
    pi0[iz] = pi0[iz - 1] - (GEE/th0[iz - 1])*dz;
    p0[iz] = pfrompi(pi0[iz]);
    if (iz < nz) {
      z = zcarray[iz];

      /* the centered pi lags one grid level -- why are we doing this? */
      cpi0 = 0.5*(pi0[iz] + pi0[iz - 1]);
      grate = (z < tpause) ? stab*gmoist(th0[iz - 1],cpi0) : 12.;
      th0[iz] = th0[iz - 1] + grate*dz;
    }
  }

  /* get the centered density and centered pres from the edged pressure using
   * the hydrostatic equation */
  for (iz = 0; iz < nz; iz++) {
    rho0[iz] = -(p0[iz + 1] - p0[iz])/(GEE*dz);

    /* also compute the mixing ratio and equivalent potential temperature */
    cpi0 = 0.5*(pi0[iz + 1] + pi0[iz]);
    rt0[iz] = rh*rs(th0[iz],cpi0);
    the0[iz] = thetaecomp(th0[iz], rt0[iz], cpi0);
    thes0[iz] = thetaescomp(th0[iz], cpi0);
    pc0[iz] = 0.5*(p0[iz + 1] + p0[iz]);
  }

  /* free space */
  fbuf(p0);
  fbuf(pi0);
}

/* hydrostatic2 -- Compute initial hydrostatic profiles of density and
 * equivalent potential temperature, given potential temperature and
 * mixing ratio.  Don't recompute grids or theta profile like hydrostatic.
 */
static void hydrostatic2(int nz, double dz, double psurf, double *th0,
			 double *the0, double *thes0, double *rho0,
			 double *rt0, double *pc0, double *cpi0)
{
  int iz;
  double *p0, *pi0;
  double z;

  /* allocate space */
  p0 = doublebuff(nz + 1);
  pi0 = doublebuff(nz + 1);

  /* integrate the hydrostatic equation to get pressure on cell edges
   * and density on cell centers */
  pi0[0] = pifromp(psurf);
  p0[0] = pfrompi(pi0[0]);
  for (iz = 1; iz <= nz; iz++) {
    pi0[iz] = pi0[iz - 1] - (GEE/th0[iz - 1])*dz;
    p0[iz] = pfrompi(pi0[iz]);
  }

  /* get the centered density and pressure from the edged pressure using
   * the hydrostatic equation */
  for (iz = 0; iz < nz; iz++) {
    rho0[iz] = -(p0[iz + 1] - p0[iz])/(GEE*dz);
    pc0[iz] = 0.5*(p0[iz + 1] + p0[iz]);

    /* also compute the equivalent potential temperature */
    cpi0[iz] = 0.5*(pi0[iz + 1] + pi0[iz]);
    the0[iz] = thetaecomp(th0[iz], rt0[iz], cpi0[iz]);
    thes0[iz] = thetaescomp(th0[iz], cpi0[iz]);
  }

  /* free space */
  fbuf(p0);
  fbuf(pi0);
}



/* this is the entry point */
main(int argc, char *argv[])
{
  int doprofile;
  char basename[121], profname[121];
  char startfilename[121];
  long nz, iz, it, jt, nt, mt, izwtg, frad, land;
  double dz, dt, rh, tpause, stab, vx1, vx2, vx3, vy1,vy2;
  double drelax, trelax, rtmul, zscaled;
  double radcool, radbrk, wtgbase, minthgrad, thprime, sst0, lfrac;
  double lamtemp, lamsea, lamwater, albedo;
  double z, time, tsurf, perturbfactor, zscale, rhperturb, thetagrad;
  double estep, cvc, cvs, cve, srh, wtrain, wtsnow, theslop;
  double pscale, pbltop, cdrag, wscale, cld, cfract;
  double rtdeficit;
  double *rho, *th, *the, *thes, *rt, *vx, *vy, *p, *pi, *rsat, *irflux;
  double *rho0, *th0, *the0, *thes0, *rt0, *vx0, *vy0;
  double *theflux, *rtflux, *vxflux, *vyflux, *rain, *srain, *erain;
  double *convthrot, *fluxthrot, *cvcaug, *ctopdeep, *ctopshallow;
  double *stemp, *swater;
  double *sthe, *srt, *sth, *sthe_cu, *sth_par, *srt_cu;
  double *svx_cu, *svy_cu, *zcarray, *dzarray, *vzwtg, *dvzwtgdz;
  double *dtemp, *adjmask, *svx, *svy, *srad;

  /* these added to do heat/moist experiments a la Kuang 2010 */
  int gmode;
  double thm1, thm2, dtempdt, dqdt, dp, finetime;
  double *basis, *dtempdt1, *dqdt1;

  /* usage */
  if (argc != 2) {
    fprintf(stderr,"Usage: wtg start_file\n");
    exit(1);
  }

  /* read start file */
  strcpy(startfilename,argv[argc - 1]);
  startfile = fopen(startfilename,"r");
  getsvar("basename",basename);
  getsvar("profname",profname);
  doprofile = strcmp("-",profname);

  /* the following vars are only important if profname ="-" */
  rh = getfvar("rh");
  tpause = getfvar("tpause"); /* used in wtg and frad */
  stab = getfvar("stab");
  vx1 = getfvar("vx1");
  vx2 = getfvar("vx2");
  vx3 = getfvar("vx3");
  vy1 = getfvar("vy1");
  vy2 = getfvar("vy2");

  /* back to variables not dependent on profname */
  nz = getivar("nz");
  nt = getivar("nt");
  mt = getivar("mt");
  dz = getfvar("dz");
  dt = getfvar("dt");
  thm1 = getfvar("thm1");
  thm2 = getfvar("thm2");
  dtempdt = getfvar("dtempdt");
  dqdt = getfvar("dqdt");
  gmode = getivar("gmode");
  trelax = getfvar("trelax");
  drelax = getfvar("drelax");
  rtmul = getfvar("rtmul");
  wtgbase = getfvar("wtgbase");
  minthgrad = getfvar("minthgrad");
  sst0 = getfvar("sst0");
  lfrac = getfvar("lfrac");
  land = lfrac > 0.5 ? 1 : 0;
  lamtemp = getfvar("lamtemp");
  lamsea = getfvar("lamsea");
  lamwater = getfvar("lamwater");
  albedo = getfvar("albedo");
  radcool = getfvar("radcool");
  radbrk = getfvar("radbrk");
  estep = getfvar("estep");
  cvc = getfvar("cvc");
  cvs = getfvar("cvs");
  cve = getfvar("cve");
  srh = getfvar("srh");
  wtrain = getfvar("wtrain");
  wtsnow = getfvar("wtsnow");
  theslop = getfvar("theslop");
  pscale = getfvar("pscale");
  pbltop = getfvar("pbltop");
  cdrag = getfvar("cdrag");
  wscale = getfvar("wscale");
  frad = getivar("frad");
  cld = getfvar("cld");
  cfract = getfvar("cfract");
  close(startfile);

  /* get profile data if needed */
  if (doprofile) readprofile(nz, profname);

  /* allocate space */
  rho = doublebuff(nz);
  p = doublebuff(nz);
  th = doublebuff(nz);
  the = doublebuff(nz);
  thes = doublebuff(nz);
  rt = doublebuff(nz);
  th0 = doublebuff(nz);
  the0 = doublebuff(nz);
  thes0 = doublebuff(nz);
  rho0 = doublebuff(nz);
  rt0 = doublebuff(nz);
  vx0 = doublebuff(nz);
  vy0 = doublebuff(nz);
  vx = doublebuff(nz);
  vy = doublebuff(nz);
  sthe = doublebuff(nz);
  srt = doublebuff(nz);
  sth = doublebuff(nz);
  vzwtg = doublebuff(nz);
  dvzwtgdz = doublebuff(nz);
  dtemp = doublebuff(nz);
  adjmask = doublebuff(nz);
  svx = doublebuff(nz);
  svy = doublebuff(nz);
  srad = doublebuff(nz);
  rsat = doublebuff(nz);
  irflux = doublebuff(nz);
  pi = doublebuff(nz);
  sthe_cu = doublebuff(nz);
  sth_par = doublebuff(nz);
  srt_cu = doublebuff(nz);
  svx_cu = doublebuff(nz);
  svy_cu = doublebuff(nz);
  dzarray = doublebuff(nz);
  zcarray = doublebuff(nz);
  theflux = doublebuff(1);
  rtflux = doublebuff(1);
  vxflux = doublebuff(1);
  vyflux = doublebuff(1);
  rain = doublebuff(1);
  srain = doublebuff(1);
  erain = doublebuff(1);
  convthrot = doublebuff(1);
  fluxthrot = doublebuff(1);
  cvcaug = doublebuff(1);
  ctopdeep = doublebuff(1);
  ctopshallow = doublebuff(1);
  stemp = doublebuff(1);
  swater = doublebuff(1);
  basis = doublebuff(nz);
  dtempdt1 = doublebuff(nz);
  dqdt1 = doublebuff(nz);

  /* set up grids and adjustment mask and zero some stuff */
  for (iz = 0; iz < nz; iz++) {
    dzarray[iz] = dz;
    zcarray[iz] = (iz + 0.5)*dz;
    z = zcarray[iz];
    vzwtg[iz] = dvzwtgdz[iz] = 0.;
    adjmask[iz] = sin(3.14159*z/tpause);
          
    /* zero adjustment mask above tropopause */
    if (z > tpause && iz > 0) {
      adjmask[iz] = 0.;
    }
  }

  /* check on form of thermodynamic profile initialization */
  if (doprofile) {

    /* pre-existing profile for thermodynamic initializations */
    getprofile(0., th0, rt0, vx0, vy0);
    hydrostatic2(nz, dz, PSURF, th0, the0, thes0, rho0, rt0, p, pi);
  }
  else {

    /* no pre-existing profile -- do hydrostatic profile initialization */
    tsurf = sst0 - 2.;
    hydrostatic(nz, dz, PSURF, tsurf, rh, tpause, stab, th0, the0, thes0,
		rho0, rt0, p, zcarray);
    
    /* initialize the wind profile */
    for (iz = 0; iz < nz; iz++) {
      z = zcarray[iz];
      vx0[iz] = vx1 + vx2*z + 0.5*vx3*z*z;
      vy0[iz] = vy1 + vy2*z;
          
      /* constant wind above tropopause */
      if (z > tpause && iz > 0) {
	vx0[iz] = vx0[iz - 1];
	vy0[iz] = vy0[iz - 1];
      }
    }
  }

  /* more initializations */
  for (iz = 0; iz < nz; iz++) {
    z = zcarray[iz];

    /* initialize the evolving fields */
    th[iz] = th0[iz];
    rt[iz] = rt0[iz]*rtmul;
    pi[iz] = pifromp(p[iz]);
    the[iz] = thetaecomp(th[iz], rt[iz], pi[iz]);
    thes[iz] = thes0[iz];
    vx[iz] = vx0[iz];
    vy[iz] = vy0[iz];
    rho[iz] = rho0[iz];
    rsat[iz] = rs(th[iz],pi[iz]);

    /* zero the source and sink fields */
    sthe[iz] = srt[iz] = sth[iz] = svx[iz] = svy[iz] = srad[iz] = 0.;
    sthe_cu[iz] = sth_par[iz] = srt_cu[iz] = svx_cu[iz] = svy_cu[iz] = 0.;
    dtempdt1[iz] = dqdt1[iz] = basis[iz] = 0.;
  }

  /* initialize the surface temperature and surface water column fields */
  *stemp = sst0;
  *swater = 0.;

  /* zero time call to the parameterizations */
  diabat3(nz,land,estep,1,frad,radcool,tpause,radbrk,cvc,cvs,cve,srh,wtrain,
	  wtsnow,theslop,pscale,pbltop,cdrag,wscale,cld,cfract,lamtemp,
	  lamsea, lamwater,albedo,dt,zcarray,dzarray,th,rt,pi,vx,vy,rho,sth_par,
	  sthe_cu,srad,srt_cu,svx_cu,svy_cu,irflux,rain,srain,erain,
	  theflux,rtflux,vxflux,vyflux,convthrot,fluxthrot,cvcaug,ctopdeep,
	  ctopshallow,stemp,swater);

  /* write initial file */
  it = 0;
  time = 0.;
  docandis(basename, profname, startfilename, it, nz, dz, dt, time,
	   dzarray, zcarray, rho, th, the, thes, rt, rsat, p, vx, vy,
	   th0, the0, thes0, rho0, rt0, vx0, vy0, sthe, srt, sth,
	   sthe_cu, sth_par, srt_cu, irflux, svx_cu, svy_cu, pi, vzwtg,
	   dvzwtgdz, adjmask, svx, svy, srad, theflux, rtflux,
	   vxflux, vyflux, rain, srain, erain, convthrot,
	   fluxthrot, cvcaug, ctopdeep, ctopshallow, stemp, swater,
	   basis, dtempdt1, dqdt1);

  /* major loop */
  for (it = 1; it <= nt; it++) {
    time = it*mt*dt;

    /* minor loop */
    for (jt = 0; jt < mt; jt++) {
      
      /* add a finer time variable to do time checking */
      finetime = time + jt*dt;

      /* modify the reference profile in time-dependent fashion and
       * fix hydrostatic balance at each step
       */
      if (doprofile) {
	getprofile(time, th0, rt0, vx0, vy0);
	hydrostatic2(nz, dz, PSURF, th0, the0, thes0, rho0, rt0, p, pi);
      }
      
      /* compute the local sink terms -- theta-e is forced so as
       * to drive theta toward the desired profile; vx and vy
       * are forced to drive these quantities toward a specified
       * profile; rt is forced by assumed entrainment and
       * vertical advection consistent with theta source
       */
      izwtg = wtgbase/dz + 0.5;
      if (izwtg < 1) izwtg = 1;
      if (izwtg > nz - 1) izwtg = nz - 1;
      for (iz = 0; iz < nz; iz++) {

	/* heating term */
	thprime = th[iz] - th0[iz];
	sth[iz] = trelax*adjmask[iz]*thprime;

	/* compute wtg vertical velocity using upstream differencing */
	if (iz > 0 && iz < nz - 1) {
	  if (sth[iz] > 0.) {
	    thetagrad = (th[iz] - th[iz - 1])/dz;
	  }
	  else {
	    thetagrad = (th[iz + 1] - th[iz])/dz;
	  }
	  if (thetagrad < minthgrad) thetagrad = minthgrad;
	  vzwtg[iz] = sth[iz]/thetagrad;
	}
      }

      /* linearly interpolate from zero at top and bottom */
      vzwtg[0] = 0.3333*vzwtg[1];
      vzwtg[nz - 1] = 0.3333*vzwtg[nz - 2];

      /* fix wtg vertical velocity in boundary layer by linear
       * interpolation between the surface and the assumed boundary
       * layer top */
      for (iz = 0; iz < izwtg; iz++) {
	vzwtg[iz] = (zcarray[iz]*vzwtg[izwtg])/zcarray[izwtg];
      }

      /* compute divergence term, extrapolating to endpoints */
      for (iz = 1; iz < nz - 1; iz++) {
	dvzwtgdz[iz] = (rho0[iz + 1]*vzwtg[iz + 1] -
			rho0[iz - 1]*vzwtg[iz - 1])/(2.*rho0[iz]*dz);
      }
      dvzwtgdz[0] = 2.*dvzwtgdz[1] - dvzwtgdz[2];
      dvzwtgdz[nz - 1] = 2.*dvzwtgdz[nz - 2] - dvzwtgdz[nz - 3];

      /* smooth the divergence */
      for (iz = 1; iz < nz - 1; iz++) {
	dtemp[iz] = 0.5*dvzwtgdz[iz] + 0.25*(dvzwtgdz[iz + 1] +
					     dvzwtgdz[iz - 1]);
      }
      for (iz = 1; iz < nz - 1; iz++) dvzwtgdz[iz] = dtemp[iz];

      /* moisture and theta-e sink terms -- first zero */
      for (iz = 0; iz < nz; iz++) {
	sthe[iz] = srt[iz] = 0.;
	dtempdt1[iz] = dqdt1[iz] = basis[iz] = 0.;
      }

      /* entrainment */
      for (iz = 0; iz < nz; iz++) {
	if (dvzwtgdz[iz] > 0.) {
	  srt[iz] += (rt[iz] - rt0[iz])*dvzwtgdz[iz];
	  sthe[iz] += (the[iz] - the0[iz])*dvzwtgdz[iz];
	}
      }

      /* forced heating, and moistening implemented through theta-e 
	 and total cloud water sink terms since they don't pass thru 
	 diabat3. Since it is a sink term, we subtract
	 it here, so the user gets a properly signed anomaly! 
	 This to implement Z. Kuang's linear response function analysis
	 [dt] = ks
	 [dtempdt] = K/day
	 [dqdt] = g/kg/day
	 [sthe] = K/ks
	 [srt] = kg/kg/ks
	 [dtempdt1] = K/day
	 [dqdt1] = g/kg/day
      */ 
      if (thm2 > thm1 && finetime >= thm1 && time < thm2) {
	for (iz = 0; iz < nz; iz++) {

	  /* use proper mode (gmode=0 is the special surface mode) */
	  if (gmode == 0) {
	    dp=3000.;
	    basis[iz] = exp(-pow((p[iz] - p[0])/dp,2));
	  }
	  else {
	    dp=7500.;
	    basis[iz] = exp(-pow((p[iz] - p[0] + (gmode - 0.5)*dp)/dp,2));
	  }

	  /* halve the perturbations above 500hPa and again above 250hPa */
	  if (p[iz] < 50000.) {
	    dtempdt1[iz] = dtempdt*basis[iz]/2.;
	    dqdt1[iz] = dqdt*basis[iz]/2.;
	    if (p[iz] < 25000.) {
	      dtempdt1[iz] = dtempdt*basis[iz]/4.;
	      dqdt1[iz] = dqdt*basis[iz]/4.;
	    }
	  }
	  else {
	    dtempdt1[iz] = dtempdt*basis[iz];
	    dqdt1[iz] = dqdt*basis[iz];
	  }
	  
	  /* apply the heating/moistening tendencies */
	  sthe[iz] -= dtempdt1[iz]*(1./86.4);
	  srt[iz] -= dqdt1[iz]*(1./86.4)*(1./1000.);
	}
      }

      /* vertical advection -- upstream diff */
      for (iz = 1; iz < nz - 1; iz++) {
	if (vzwtg[iz] > 0.) {
	  srt[iz] += vzwtg[iz]*(rt[iz] - rt[iz - 1])/dz;
	  sthe[iz] += vzwtg[iz]*(the[iz] - the[iz - 1])/dz;
	}
	else {
	  srt[iz] += vzwtg[iz]*(rt[iz + 1] - rt[iz])/dz;
	  sthe[iz] += vzwtg[iz]*(the[iz + 1] - the[iz])/dz;
	}
      }

      /* finish setup for call of cumulus parameterization */
      for (iz = 0; iz < nz; iz++) {
	svx[iz] = (vx[iz] - vx0[iz])*drelax;
	svy[iz] = (vy[iz] - vy0[iz])*drelax;
	rsat[iz] = rs(th[iz],pi[iz]);
	the[iz] = thetaecomp2(th[iz],rt[iz],rsat[iz]);
      }

      /* call the parameterizations */
      diabat3(nz,land,estep,1,frad,radcool,tpause,radbrk,cvc,cvs,cve,srh,
	      wtrain,wtsnow,theslop,pscale,pbltop,cdrag,wscale,
	      cld,cfract,lamtemp,lamsea,lamwater,albedo,dt,zcarray,dzarray,
	      th,rt,pi,vx,vy,rho,sth_par,sthe_cu,srad,srt_cu,svx_cu,svy_cu,
	      irflux,rain,srain,erain,theflux,rtflux,vxflux,vyflux,
	      convthrot,fluxthrot,cvcaug,ctopdeep,ctopshallow,stemp,swater);

      /* update the profiles and finish diagnostics */
      rtdeficit = 0.;
      for (iz = nz - 1; iz >= 0; iz--) {
	sth_par[iz] = th[iz]*((sthe_cu[iz] + srad[iz])/the[iz] -
			     ALFA*srt_cu[iz]);
	the[iz] += (sthe_cu[iz] + srad[iz] - sthe[iz])*dt;
	rt[iz] += (srt_cu[iz] - srt[iz])*dt - rtdeficit;
	
	/* output only the convective response, not the prescribed tend. */
	/* these outputs are in K/day, g/kg/day */
	dtempdt1[iz] = (sthe_cu[iz] + srad[iz] - sthe[iz] + \
			dtempdt1[iz]*(1./86.4))*86.4;
	dqdt1[iz] = (srt_cu[iz] - srt[iz] + dqdt1[iz]*(1./86.4)*(1./1000) \
		     - rtdeficit/dt)*86.4*1000.;

	/* propagate any negative rt downward */
	if (rt[iz] < 0.) {
	  rtdeficit = - rt[iz];
	  rt[iz] = 0.;
	}
	vx[iz] += (svx_cu[iz] - svx[iz])*dt;
	vy[iz] += (svy_cu[iz] - svy[iz])*dt;
	th[iz] = thetacomp(the[iz],rt[iz],pi[iz]);
	rsat[iz] = rs(th[iz],pi[iz]);
      }

      /* redo hydrostatic calculation -- this redoes rho, the, thes,
       * p, and pi, starting with th
       */
      hydrostatic2(nz, dz, PSURF, th, the, thes, rho, rt, p, pi);
    }
    
    /* write an output file */
    docandis(basename, profname, startfilename, it, nz, dz, dt, time,
	     dzarray, zcarray, rho, th, the, thes, rt, rsat, p, vx, vy,
	     th0, the0, thes0, rho0, rt0, vx0, vy0, sthe, srt, sth, sthe_cu, 
	     sth_par, srt_cu, irflux, svx_cu, svy_cu, pi, vzwtg, dvzwtgdz,
	     adjmask, svx, svy, srad, theflux, rtflux, vxflux, vyflux,
	     rain, srain, erain, convthrot, fluxthrot, cvcaug,
	     ctopdeep, ctopshallow, stemp, swater, basis, dtempdt1, dqdt1);
  }

  exit(0);
}
