/* diabat3c -- Version 3 of Raymond convective/radiative package.
 *
 * Note: This subroutine provides a parameterization for **all**
 * diabatic effects, including latent heat release in explicit
 * vertical motion and in hard adjustment due to dry static
 * instability.  Don't include any separate explicit diabatic
 * effects or you will be double counting them!
 *
 * The model accepts input fields zc thru w in a single column at cell-
 * centered levels defined by zc.  The output profiles sth thru sv
 * have the same structure.  The values of dz for each cell
 * should be consistent with zc.  The levels must be ordered from
 * lowest to highest in height.  Tendencies for potential temperature,
 * mixing ratio, and the two horizontal wind components are returned
 * at these levels.  Note that the unit ``ks'' is kiloseconds.
 *
 */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "diabat3.h"
#include "wallclock.h"



/* safeexp -- Do our own computation of the exponential function --
 * this results from the sporadic long times required to compute
 * the exponential function using the standard GNU math library on
 * 64 bit systems (AMD, Intel)
 */
#define E 2.7182818284590451
#define ETABLESIZE 100001
static TFLOAT safeexp(TFLOAT x)
{
  static TFLOAT xx, result, table[ETABLESIZE];
  static int begin = 1, i;

  if (begin) {
    for (i = 0; i < ETABLESIZE; i++) {
      table[i] = exp(((double)i)/((double)(ETABLESIZE)));
    }
    begin = 0;
  }

  xx = x;
  result = 1.;
  while (xx >= 0.) {
    xx -= 1.;
    result *= E;
  }
  while (xx < 0.) {
    xx += 1.;
    result /= E;
  }
  i = xx*ETABLESIZE;
  if (i < 0) i = 0;
  if (i >= ETABLESIZE) i = ETABLESIZE - 1;
  result *= table[i];
  return(result);
}


/* floatbuff -- This routine uses malloc to allocate float buffer
 * space.  The argument nelem defines the number of floating
 * point elements.  It returns a float pointer to the space.  If malloc
 * fails, getbuff dies with an error message.
 */
static TFLOAT *floatbuff(TLONG nelem)
{
  TFLOAT *point;
  if ((point = (TFLOAT*)malloc((unsigned)(sizeof(TFLOAT)*nelem)))
      == (TFLOAT*)NULL) {
    fprintf(stderr,"floatbuff: can't get %d words\n",nelem);
    exit(1);
  }
  return(point);
}

/* fbuf -- Free a buffer obtained with floatbuff */
static void fbuf(TFLOAT *ptr)
{
  free((char*)ptr);
}

/* surfextrap -- Extrapolate a field to the surface */
static TFLOAT surfextrap(TFLOAT *fld, TFLOAT *dz)
{
  return(fld[0] + (fld[0] - fld[1])*dz[0]/(dz[1] + dz[0]));
}





/* cuparam -- The convective parameterization */

/* constants for the cumulus parameterization */
#define FTHROTSLOP 0.5
#define MAXFRAC 0.3
#define DTMAX 1.0
#define PI 3.14159
#define DEEP 1
#define SHALLOW 2
#define ALPHASHALLOW (-1.)
#define ALPHADEEP (0.)

/* pblfld -- PBL mean value of field -- this is approximated by
 * an average with an exponentially decaying weight with efolding
 * equal to the PBL depth -- it is done this way to avoid shocks
 * caused by index flipping
 */
static TFLOAT pblfld(TLONG nz, TFLOAT pbltop, TFLOAT *zc, TFLOAT *fld)
{
  TLONG iz;
  TFLOAT fldval, normval, weight, zscaled;

  fldval = 0.;
  normval = 0.;
  for (iz = 0; iz < nz; iz++) {
    zscaled = zc[iz]/pbltop;
    weight = exp(-zscaled*zscaled);
    fldval += weight*fld[iz];
    normval += weight;
  }
  fldval /= normval;
  return(fldval);
}

/* thresh -- compute threshold value of theta-e for convection -- this
 * is roughly equal to the saturated equivalent potential temperature
 * between one and two PBL depths -- computed as a weighted average
 * with a weight maximizing in the specified range
 */
static TFLOAT thresh(TLONG nz, TFLOAT pbltop, TFLOAT *zc,
		     TFLOAT *th, TFLOAT *rt, TFLOAT *pi)
{
  TFLOAT thesval, normval, weight, zscaled, zdev;
  TINT iz;

  /* get weighted average of saturation theta-e in region above pbl */
  thesval = 0.;
  normval = 0.;
  for (iz = 0; iz < nz; iz++) {
    zscaled = zc[iz]/pbltop;
    zdev = zscaled - 1.5;
    weight = exp(-zdev*zdev);
    thesval += weight*thetaescomp(th[iz],pi[iz]);
    normval += weight;
  }
  thesval /= normval;
  return(thesval);
}




/* wt -- compute particle terminal velocity */
static TFLOAT wt(TFLOAT th, TFLOAT pi, TFLOAT wtrain, TFLOAT wtsnow)
{
  TFLOAT temp;
  temp = th*pi/CP;
  return(temp > FREEZING ? wtrain : wtsnow);
}

/* throtcomp -- evaluate the throttle function */
static TFLOAT throtcomp(TFLOAT value, TFLOAT slop)
{
  TFLOAT tvalue;
  tvalue = 0.5 + value/slop;
  if (tvalue < 0.) tvalue = 0.;
  if (tvalue > 1.) tvalue = 1.;
  return(tvalue);
}

/* buoy -- compute buoyancy of a surface parcel at an arbitrary level */
static TFLOAT buoy(TLONG level, TFLOAT thesurf, TFLOAT rtsurf,
		   TFLOAT *pi, TFLOAT *th)
{
  return(thetacomp2(thesurf,rtsurf,pi[level]) - th[level]);
}

/* vfract -- compute valid fraction of a partially full cell */
static TFLOAT vfract(TLONG iz, TLONG top, TFLOAT ztfract)
{
  TFLOAT validfract;
  if (iz >= top) {
    validfract = 0.;
  }
  else if (iz == top - 1) {
    validfract = ztfract;
  }
  else {
    validfract = 1.;
  }
  return(validfract);
}

/* sources -- do the real work of the convective parameterization */

static void sources(TLONG nz, TLONG mode, TFLOAT dt, TFLOAT alpha,
		    TFLOAT estep, TFLOAT cvc, TFLOAT cvs,
		    TFLOAT cve, TFLOAT srh, TFLOAT wtrain,
		    TFLOAT wtsnow, TFLOAT pscale,
		    TFLOAT pbltop, TFLOAT *zc, TFLOAT *dz, TFLOAT *pi,
		    TFLOAT *th, TFLOAT *the, TFLOAT *rt, TFLOAT *u,
		    TFLOAT *v, TFLOAT *rho, TFLOAT *rsat, TFLOAT eflux,
		    TFLOAT rflux, TFLOAT uflux, TFLOAT vflux,
		    TFLOAT *sthe, TFLOAT *srt, TFLOAT *su,
		    TFLOAT *sv, TFLOAT *rain, TFLOAT *srain,
		    TFLOAT *erain, TFLOAT *cvcaug, TFLOAT *ctop)
{
  TLONG iz,jz;               /* z indices */
  TLONG top;                 /* index just above cloud top */
  TFLOAT stratx,evapx,totx;  /* for precipitation computation */
  TFLOAT satvalue,srainval;  /* for precipitation computation */
  TFLOAT sprecip,eprecip,precip;  /* precipitation components */
  TFLOAT cvcactual;          /* actual value of cvc */
  TFLOAT validfract;         /* fraction of cell occupied by cloud */
  TFLOAT thegrad;            /* vertical gradient of the */
  TFLOAT emean,rmean,umean,vmean,lmean;  /* stuff for mean value comps */
  TFLOAT ztfract,buoy1,buoy2; /* for fractional high top cell calc */
  TFLOAT curateint;          /* mass-weighted integral of curate(z) */
  TFLOAT mass;               /* mass per unit area in a layer */
  static TFLOAT *linear;     /* zero-mean linear profile */
  static TFLOAT *curate;     /* weighting profile for mixing */
  TFLOAT ztop;               /* top of convection */
  TFLOAT thepbl,rtpbl;       /* pbl values of fields */
  static int begin = 1;

  /* allocate space only once */
  if (begin) {
    curate = floatbuff(nz);
    linear = floatbuff(nz);
    begin = 0;
  }

  /* compute the convective top for each mode */
  if (mode == DEEP) {  /* deep mode */

    /* compute the top of the convective layer as the level above the first
     * positively buoyant layer marching from the top down -- the top layer
     * used in all the computations is actually top - 1, i. e., the last
     * positively buoyant layer
     */
    thepbl = pblfld(nz,pbltop,zc,the);
    rtpbl = pblfld(nz,pbltop,zc,rt);
    top = nz - 2;
    while (top > 0 && buoy(top,thepbl,rtpbl,pi,th) < 0.) {
      top--;
    }
    
    /* note that top is always > 0 */
    top++;

    /* compute the fractional position of neutral level in high top layer */
    buoy1 = buoy(top,the[0],rt[0],pi,th);
    buoy2 = buoy(top - 1,the[0],rt[0],pi,th);
    if (top == 1) {
      ztfract = 1.;
    }
    else if (buoy2 - buoy1 < 1.e-10) {
      ztfract = 0.5;
    }
    else {
      ztfract = buoy2/(buoy2 - buoy1);
    }
    if (ztfract > 1.) ztfract = 1.;
    if (ztfract < 0.) ztfract = 0.;

    /* convective top for deep convection */
    ztop = zc[top - 1] + ztfract*(dz[top - 1] + dz[top])/2.;
    if (ztop < pbltop) ztop = pbltop;
  }
  else {  /* pbl mode -- top is specified */
    ztop = pbltop;
    top = 1;
    while (ztop > zc[top]) top++;
    ztfract = (ztop - zc[top - 1])/(zc[top] - zc[top - 1]);
    if (ztfract > 1.) ztfract = 1.;
    if (ztfract < 0.) ztfract = 0.;
    ztop = zc[top - 1] + ztfract*(dz[top - 1] + dz[top])/2.;
  }
  *ctop = ztop;

  /* compute initial linear structure function */
  for (iz = 0; iz < nz; iz++) {
    linear[iz] = zc[iz];
  }

  /* rate limit mixing to avoid rates higher than the time step accommodates */
  cvcactual = cvc;
  if (cvcactual*DTMAX > MAXFRAC) cvcactual = MAXFRAC/DTMAX;

  /* compute means */
  curateint = emean = rmean = umean = vmean = lmean = 0.;
  for (iz = 0; iz < nz; iz++) {

    /* compute valid fraction of cell to include */
    validfract = vfract(iz,top,ztfract);

    /* compute height dependent convective rate */
    curate[iz] = cvcactual*(1. + alpha*zc[iz]/ztop)*validfract;
    if (curate[iz] < 0.) curate[iz] = 0.;

    /* compute the means */
    mass = dz[iz]*rho[iz];
    curateint += curate[iz]*mass;
    emean += curate[iz]*the[iz]*mass;
    rmean += curate[iz]*rt[iz]*mass;
    umean += curate[iz]*u[iz]*mass;
    vmean += curate[iz]*v[iz]*mass;
    lmean += curate[iz]*linear[iz]*mass;
  }
  emean /= curateint;
  rmean /= curateint;
  umean /= curateint;
  vmean /= curateint;
  lmean /= curateint;

  /* give linear weighted zero mean over the depth of the convection
   * and compute the RMS variance of theta-e
   */
  for (iz = 0; iz < nz; iz++) {
    linear[iz] -= lmean;
  }

  /* compute precip and all source terms -- march down from highest level */
  iz = nz - 1;
  precip = sprecip = eprecip = 0.;
  while (iz >= 0) {

    /* stratiform rain production -- can't occur if dthe/dz < 0 */
    if (iz >= top - 2) {
      thegrad = 10.;
    }
    else {
      thegrad = (the[iz + 1] - the[iz])/(zc[iz + 1] - zc[iz]);
    }
    srainval = rt[iz] - srh*rsat[iz];
    stratx = srainval > 0. ? 
      srainval*cvs*throtcomp(thegrad - 1.,2.)*
      throtcomp(zc[iz] - pscale,pscale) : 0.;

    /* evaporation of rain */
    satvalue = rt[iz] - rsat[iz];
    evapx = cve*satvalue*precip/(rho[iz]*wt(th[iz],pi[iz],wtrain,wtsnow));
    if (evapx > 0.) evapx = 0.;

    /* accumulate the precipitation components by downward integration */
    mass = rho[iz]*dz[iz];
    sprecip += stratx*mass;
    eprecip += evapx*mass;
    precip = sprecip + eprecip;

    /* don't let cumulative precip become negative due to excessive evap */
    if (precip < 0.) {
      eprecip -= precip;
      evapx -= precip/mass;
      precip = 0.;
    }
    totx = stratx + evapx;

    /* source terms */
    sthe[iz] = (emean + estep*linear[iz] - the[iz] +
		eflux/curateint)*curate[iz];
    srt[iz] = (rmean - rt[iz] + rflux/curateint)*curate[iz] - totx;
    su[iz] = (umean - u[iz] + uflux/curateint)*curate[iz];
    sv[iz] = (vmean - v[iz] + vflux/curateint)*curate[iz];
    
    /* decrement counter */
    iz--;
  }

  /* return the precipitation components */
  *rain = precip;
  *srain = sprecip;
  *erain = eprecip;

  /* return the augmented convective mixing rate */
  *cvcaug = cvcactual;
}

/* the main routine for the cumulus parameterization */
static void cuparam(TLONG nz, TLONG estep, TFLOAT cvc, TFLOAT cvs,
		    TFLOAT cve, TFLOAT srh, TFLOAT wtrain, TFLOAT wtsnow,
		    TFLOAT theslop, TFLOAT pscale, TFLOAT pbltop,
		    TFLOAT cdrag, TFLOAT wscale, TFLOAT sst, TFLOAT landrh,
		    TFLOAT dt, TFLOAT *zc, TFLOAT *dz, TFLOAT *th,
		    TFLOAT *the, TFLOAT *rt, TFLOAT *pi, TFLOAT *u, TFLOAT *v,
		    TFLOAT *rho, TFLOAT *rsat, TFLOAT *sthe, TFLOAT *srt,
		    TFLOAT *su, TFLOAT *sv, TFLOAT *rain, TFLOAT *srain,
		    TFLOAT *erain, TFLOAT *eflux, TFLOAT *rflux,
		    TFLOAT *uflux, TFLOAT *vflux, TFLOAT *convthrot,
		    TFLOAT *fluxthrot, TFLOAT *cvcaug, TFLOAT *ctopdeep,
		    TFLOAT *ctopshallow)
{
  TLONG iz;                   /* looping index */
  TFLOAT ewind;               /* effective wind for fluxes */
  TFLOAT surfpi;              /* extrapolated Exner function at surface */
  TFLOAT surftemp;            /* extrapolated temperature at surface */
  TFLOAT surfthe;             /* extrapolated thetae at surface */
  TFLOAT surfrt;              /* extrapolated mixing ratio at surface */
  TFLOAT surfth;              /* extrapolated potential temperature at surf */
  TFLOAT surfu,surfv;         /* extrapolated wind at surface */
  TFLOAT sstheta;             /* potential temperature of sea surface */
  TFLOAT ssthe;               /* equivalent potential temperature of surface */
  TFLOAT ssrt;                /* mixing ratio of surface */
  TFLOAT fthrottle;           /* throttle on fluxes due to sea-air temp diff */
  static TFLOAT *sthe1,*srt1,*su1,*sv1;  /* deep forcing terms */
  static TFLOAT *sthe2,*srt2,*su2,*sv2;  /* pbl forcing terms */
  TFLOAT rain1,srain1,erain1;  /* deep rain */
  TFLOAT rain2,srain2,erain2;  /* shallow rain */
  TFLOAT cvcaug1,cvcaug2;     /* augmented mixing rate */
  TFLOAT throttle,cthrottle;  /* deep convective throttle and complement */
  TFLOAT srtcorr,precipcorr;  /* correct for negative rt generation */
  TFLOAT srtint;
  static int begin = 1;

  /* check size of dt */
  if (dt > DTMAX) {
    fprintf(stderr,"diabat3: Sorry, must have dt < %f\n",DTMAX);
    exit(1);
  }

  /* allocate space only once */
  if (begin) {
    sthe1 = floatbuff(nz);
    srt1 = floatbuff(nz);
    su1 = floatbuff(nz);
    sv1 = floatbuff(nz);
    sthe2 = floatbuff(nz);
    srt2 = floatbuff(nz);
    su2 = floatbuff(nz);
    sv2 = floatbuff(nz);
    begin = 0;
  }

  /* compute surface fluxes -- currently uses same drag coeff for
   * thermodynamics and wind
   */

  /* extrapolate things to the surface */
  surfpi = surfextrap(pi,dz);
  surfth = surfextrap(th,dz);
  surfthe = surfextrap(the,dz);
  surfrt = surfextrap(rt,dz);
  surfu = surfextrap(u,dz);
  surfv = surfextrap(v,dz);
  surftemp = surfth*surfpi/CP;
  sstheta = sst*CP/surfpi;       /* surface theta */

  /* Land surfaces are distinguished from sea surfaces by the surface
   * mixing ratio.  In the case of pure dry land (landrh = 0), the
   * surface mixing ratio is set equal to the model mixing ratio at
   * the lowest model level.  For pure ocean it is set to the saturation
   * mixing ratio at the surface temperature and pressure (landrh = 1).
   * For intermediate cases, landrh is set to the appropriate value.
   */
  ssrt = rs(sstheta,surfpi)*landrh + surfrt*(1. - landrh);   /* surf mix rat */
  ssthe = thetaecomp(sstheta,ssrt,surfpi);  /* surface thetae */

  /* this formula for the effective wind keeps the fluxes from going to
   * zero at zero real wind
   */
  ewind = pow(surfu*surfu + surfv*surfv + wscale*wscale,0.5);

  /* this throttle function turns off surface fluxes if the air temperature
   * at the surface is much in excess of the sea surface temperature
   */
  fthrottle = throtcomp(sst - surftemp,FTHROTSLOP);
  
  /* surface fluxes are done via a bulk formula -- the fraction of
   * land controls the surface evaporation flux
   */
  *eflux = fthrottle*rho[0]*ewind*cdrag*(ssthe - surfthe);
  *rflux = fthrottle*rho[0]*ewind*cdrag*(ssrt - surfrt);
  
  /* kill negative thermodynamic fluxes */
  if (*eflux < 0.) *eflux = 0.;
  if (*rflux < 0.) *rflux = 0.;

  /* momentum fluxes */
  *uflux = -fthrottle*rho[0]*ewind*cdrag*surfu;
  *vflux = -fthrottle*rho[0]*ewind*cdrag*surfv;

  /* compute source terms for convection -- mode 1 --> deep, mode2 --> pbl */
  sources(nz,DEEP,dt,ALPHADEEP,estep,cvc,cvs,cve,srh,wtrain,wtsnow,
	  pscale,pbltop,zc,dz,pi,th,the,rt,u,v,rho,rsat,*eflux,
	  *rflux,*uflux,*vflux,sthe1,srt1,su1,sv1,&rain1,&srain1,
	  &erain1,&cvcaug1,ctopdeep);
  sources(nz,SHALLOW,dt,ALPHASHALLOW,estep,cvc,cvs,cve,srh,wtrain,wtsnow,
	  pscale,pbltop,zc,dz,pi,th,the,rt,u,v,rho,rsat,*eflux,
	  *rflux,*uflux,*vflux,sthe2,srt2,su2,sv2,&rain2,&srain2,
	  &erain2,&cvcaug2,ctopshallow);

  /* compute convective threshold throttle function and its complement */
  throttle = throtcomp(pblfld(nz,pbltop,zc,the) -
		       thresh(nz,pbltop,zc,th,rt,pi),theslop);
  cthrottle = 1. - throttle;

  /* mix deep and pbl source terms */
  srtint = 0.;
  precipcorr = 0.;
  for (iz = 0; iz < nz; iz++) {
    sthe[iz] = throttle*sthe1[iz] + cthrottle*sthe2[iz];
    srt[iz] = throttle*srt1[iz] + cthrottle*srt2[iz];
    su[iz] = throttle*su1[iz] + cthrottle*su2[iz];
    sv[iz] = throttle*sv1[iz] + cthrottle*sv2[iz];

    /* keep srt from producing negative total water */
    if (dt > 0. && rt[iz] + srt[iz]*dt < 0.) {
      srtcorr = -srt[iz] - rt[iz]/dt;
      srt[iz] += srtcorr;
      precipcorr -= srtcorr*rho[iz]*dz[iz];
    }
    srtint += srt[iz]*rho[iz]*dz[iz];
  }
  *rain = throttle*rain1 + cthrottle*rain2 + precipcorr;
  *srain = throttle*srain1 + cthrottle*srain2;
  *erain = throttle*erain1 + cthrottle*erain2;
  *fluxthrot = fthrottle;
  *convthrot = throttle;
  *cvcaug = cvcaug1;   /* the deep mixing rate is our concern */
}





/* radiation -- Parameterization of long and shortwave radiation
 * this model is tuned by making rough comparisons with NCAR's
 * CCM2 radiation model (constructed by Jeff Kiehl and associates)
 */

/* one and only one of the following radiation options must be turned
   on (on = 1, off = 0) -- 1 channel radiation is faster, but is just
   a gray body scheme -- 8 channel radiation is the radiation scheme
   used in the Raymond-Torres (1998) paper -- J. Atmos. Sci., 55,
   1771-1790 -- and includes a crude representation of radiation-water
   vapor interactions */
#define ONECHANNEL 0
#define MULTICHANNEL 0
#define MEGACHANNEL 1

/* constants for the radiation parameterization */
#if ONECHANNEL
#define NBANDS 1
#endif

#if MULTICHANNEL
#define NBANDS 8
#endif

#if MEGACHANNEL
#define NBANDS 13
#endif

#define CO2 0
#define CONT 1
#define RTREF 0.02
#define SBCONST 5.67e-8
#define SOLARCONST 1360.

/* elimit -- limit the value to +- 75 to avoid blowing up exponential fn */
static TDOUBLE elimit(val)
TFLOAT val;
{
  TFLOAT lval;
  lval = val;
  if (lval > 75.) lval = 75.;
  if (lval < -75.) lval = -75.;
  return(lval);
}

/* zedgecomp -- extrapolate cell edge array from cell centered array */
static void zedgecomp(TLONG nz, TFLOAT *dz, TFLOAT *vcell,
		      TFLOAT *vedge)
{
  TLONG iz,jz;
  for (iz = 1; iz < nz; iz++) {
    jz = iz - 1;
    vedge[iz] = (vcell[iz]*dz[jz] + vcell[jz]*dz[iz])/(dz[iz] + dz[jz]);
  }
  vedge[0] = (vcell[0]*(dz[1] + 2.*dz[0]) - vcell[1]*dz[0])/(dz[1] + dz[0]);
  vedge[nz] = (vcell[nz - 1]*(dz[nz - 2] + 2.*dz[nz - 1]) -
	       vcell[nz - 2]*dz[nz - 1])/(dz[nz - 2] + dz[nz - 1]);
}

/* lcomp -- compute the l factor from delta tau and the b values
 * this assumes that b = sigma*T^4 is linear across the grid cell
 * -- integrate analytically and watch out for small dtau
 */
static TFLOAT lcomp(TFLOAT dtau, TFLOAT b1, TFLOAT b2)
{
  TFLOAT efact,omefact,lfactor;
  efact = safeexp(elimit(-dtau));
  omefact = 1. - efact;
  if (dtau < 0.001) {
    lfactor = 0.5*dtau*(b1 + b2);
  }
  else {
    lfactor = b1*(omefact/dtau - efact) + b2*(1. - omefact/dtau);
  }
  return(lfactor);
}

/* kcomp -- compute the kplus and kminus profiles */
static void kcomp(TLONG nz, TFLOAT *dz, TFLOAT *tau, TFLOAT *th,
		   TFLOAT *pi, TFLOAT *kplus, TFLOAT *kminus)
{
  TLONG iz,jz,mz;
  TFLOAT temp,dtau,efact;
  static TFLOAT *eth,*epi;           /* extrapolated cell edge th and pi */
  static TFLOAT *b;                  /* extrapolated cell edge sigma*T^4 */
  static int begin = 1;

  /* make space only once */
  mz = nz + 1;
  if (begin) {
    eth = floatbuff(mz);
    epi = floatbuff(mz);
    b = floatbuff(mz);
    begin = 0;
  }

  /* compute the Planck function profile on cell edges */
  zedgecomp(nz,dz,th,eth);
  zedgecomp(nz,dz,pi,epi);
  eth[0] = th[0];               /* neutral bl assumed */
  for (iz = 0; iz < mz; iz++) {
    temp = eth[iz]*epi[iz]/CP;
    b[iz] = SBCONST*temp*temp*temp*temp;
  }

  /* do the kplus profile */
  kplus[0] = 0.;
  for (iz = 1; iz < mz; iz++) {
    jz = iz - 1;
    dtau = tau[iz] - tau[jz];
    efact = safeexp(elimit(-dtau));
    kplus[iz] = efact*kplus[jz] + lcomp(dtau, b[jz], b[iz]);
    if (fabs(kplus[iz]) < 1.e-30) kplus[iz] = 0.;
  }

  /* do the kminus profile */
  kminus[nz] = 0.;
  for (iz = nz - 1; iz >= 0; iz--) {
    jz = iz + 1;
    dtau = tau[jz] - tau[iz];
    efact = safeexp(elimit(-dtau));
    kminus[iz] = efact*kminus[jz] + lcomp(dtau,b[jz],b[iz]);
    if (fabs(kminus[iz]) < 1.e-30) kminus[iz] = 0.;
  }
}

/* taucomp -- compute the path length for a band -- tau is defined on
 * cell edges and is zero at the surface
 */
static void taucomp(TLONG mz, TLONG ib, TFLOAT *dz, TFLOAT *th,
		    TFLOAT *rt, TFLOAT *pi, TFLOAT *rho, TFLOAT *rsat,
		    TFLOAT abs, TFLOAT cld, TFLOAT *tau)
{
  TLONG iz,kz;
  TFLOAT mutot,temp,pbroad,rl,rvap,mutotmax,rfract;

  /* integrate upward from surface to get tau -- iz = upper edge index,
   * kz = cell center and lower edge index
   */
  tau[0] = 0.;
  for (iz = 1; iz < mz; iz++) {
    kz = iz - 1;
    
    /* compute pressure broadening factor */
    temp = th[kz]*pi[kz]/CP;
    pbroad = pow(temp/TREF,0.5)*rho[kz]/RHOREF;

    /* compute vapor-condensate split */
    rl = rt[kz] - rsat[kz];
    if (rl < 0.) rl = 0.;
    rvap = rt[kz] - rl;
    rfract = rvap/RTREF;

    /* compute mutot in cell center at level kz for desired band */
    if (ib == CO2) {                 /* carbon dioxide bands */
      mutot = abs*pbroad;
    }
    else if (ib == CONT) {           /* water vapor continuum */
      mutot = abs*rfract*rfract;
    }
    else {                           /* water vapor bands */
      mutot = abs*pbroad*rfract;
    }

    /* add in cloud part */
    mutot += cld*rl;

    /* compute tau at the next level */
    tau[iz] = tau[kz] + rho[kz]*dz[kz]*mutot;
  }
}

/* the main radiation routine */
static void radiation(TLONG nz, TLONG nosun, TFLOAT cld, TFLOAT sst,
		      TFLOAT albedo, TFLOAT *zc, TFLOAT *dz, TFLOAT *th,
		      TFLOAT *the, TFLOAT *rt, TFLOAT *pi, TFLOAT *rho,
		      TFLOAT *rsat, TFLOAT *sthe, TFLOAT *flux)
{
  TLONG ib,iz;                /* index parameters */
  TLONG mz;                   /* number of grid edge points */
  TFLOAT scos;                /* factor for solar zenith angle */
  TFLOAT bsurf;               /* upward energy flux from surface */
  TFLOAT topefact,botefact,iplus,iminus;  /* intermediate results */
  TFLOAT normfactor;          /* normalization factor */
  TFLOAT visible;             /* visible solar radiation reaching surface */
  static TFLOAT *eflux;              /* total flux on cell edges */
  static TFLOAT *tau;                /* actual absorptivities */
  static TFLOAT *kplus,*kminus;      /* intermediate profiles */
  static int begin = 1;

  /* arrays containing per-band information about radiation model */

#if ONECHANNEL
  /* this has a single band independent of water vapor */

  /* specific absorption coefficients for each band */
  static TFLOAT abs[NBANDS] = {1.};

  /* downward flux of solar radiation for each band for overhead sun */
  static TFLOAT solar[NBANDS] = {0.};

  /* fraction of Planck-weighted spectrum occupied by each band */
  static TFLOAT fract[NBANDS] = {1.0};
#endif

#if MULTICHANNEL
  /* arrays containing per-band information about radiation model band
   * 0 is CO2 while band 1 is water vapor continuum bands 2-7 are
   * water vapor lines.  This is the radiation scheme used in Raymond
   * and Torres (1998)
   */

  /* specific absorption coefficients for each band */
  static float abs[NBANDS] = {2.,0.5,1.,10.,100.,1000.,10000.,100000.};

  /* downward flux of solar radiation for each band for overhead sun */
  static float solar[NBANDS] = {0.,160.,60.,30.,15.,7.,0.,0.};

  /* fraction of Planck-weighted spectrum occupied by each band */
  static float fract[NBANDS] = {0.11,0.20,0.19,0.16,0.13,0.10,0.07,0.04};
#endif

#if MEGACHANNEL
  /* arrays containing per-band information about radiation model band
   * 0 is CO2 while band 1 is water vapor continuum bands 2-12 are
   * water vapor lines.
   */

  /* specific absorption coefficients for each band */
  static float abs[NBANDS] = {2.,0.5,1.,3.16,10.,31.6,100.,316.,
			      1000.,3160.,10000.,31600.,100000.};

  /* downward flux of solar radiation for each band for overhead sun */
  static float solar[NBANDS] = {0.,160.,35.,26.,19.,13.,10.,6.,3.,
				0.,0.,0.,0.};

  /* fraction of Planck-weighted spectrum occupied by each band */
  static float fract[NBANDS] = {0.11,0.20,0.13,0.11,0.09,0.08,0.07,0.06,
				0.05,0.04,0.03,0.02,0.01};
#endif

  /* make space -- nz = number of cells, mz = number of cell edges */
  mz = nz + 1;
  if (begin) {
    eflux = floatbuff(mz);
    tau = floatbuff(mz);
    kplus = floatbuff(mz);
    kminus = floatbuff(mz);
    begin = 0;
  }

  /* define solar zenith angle factor and surface Planck function
   * we include the nosun flag for testing the scheme with no
   * solar input from the top
   */
  if (nosun) {
    scos = 0.;
  }
  else {
    scos = 1./3.14159;
  }
  bsurf = SBCONST*sst*sst*sst*sst;

  /* zero edge flux array */
  for (iz = 0; iz < mz; iz++) eflux[iz] = 0.;

  /* loop on bands */
  for (ib = 0; ib < NBANDS; ib++) {

    /* compute tau, the actual absorption coefficient, for this band */
    taucomp(mz,ib,dz,th,rt,pi,rho,rsat,abs[ib],cld,tau);

    /* compute the kplus and kminus profiles */
    kcomp(nz,dz,tau,th,pi,kplus,kminus);

    /* add in the flux profile for this band */
    for (iz = 0; iz < mz; iz++) {
      botefact = safeexp(elimit(-tau[iz]));
      topefact = safeexp(elimit(tau[iz] - tau[nz]));
      iplus = fract[ib]*(bsurf*botefact + kplus[iz]);
      iminus = scos*solar[ib]*topefact + fract[ib]*kminus[iz];
      eflux[iz] += (iplus - iminus);
    }
  }

  /* compute solar radiation reaching surface assuming zenith sun */
  visible = SOLARCONST*(1. - albedo);
  for (ib = 0; ib < NBANDS; ib++) {
    visible -= solar[ib];
  }

  /* compute the cell-centered radiative source and flux */
  for (iz = 0; iz < nz; iz++) {
    normfactor = the[iz]/(rho[iz]*pi[iz]*th[iz]);
    sthe[iz] = -normfactor*(eflux[iz + 1] - eflux[iz])/dz[iz];
    flux[iz] = 0.5*(eflux[iz + 1] + eflux[iz]) - scos*visible;
  }
}

/* fixed radiation code */
static void fradiation(TLONG nz, TFLOAT *zc, TFLOAT radcool, TFLOAT tpause,
	   TFLOAT radbrk, TFLOAT *sthe_rad, TFLOAT *sthe_radclr,
	   TFLOAT *irflux, TFLOAT *irfluxclr)
{
  TLONG iz;
  TFLOAT z;

  for (iz = 0; iz < nz; iz++) {
    z = zc[iz];

    /* compute the fixed radiative cooling profile */
    if (z < radbrk*tpause) {
      sthe_rad[iz] = 1.;
    }
    else if (z < tpause && z > radbrk*tpause) {
      sthe_rad[iz] = 1. - (z/tpause - radbrk)/(1. - radbrk);
    }
    else {
      sthe_rad[iz] = 0.;
    }
    sthe_rad[iz] = -sthe_rad[iz]*radcool/86.4;

    /* fill in other stuff -- ignore irflux var for now */
    sthe_radclr[iz] = sthe_rad[iz];
    irflux[iz] = irfluxclr[iz] = 0.;
  }
}





/* diabat3 -- Make all the diabatic calculations
 * nz - number of vertical grid cells
 * land - land flag (1 = land, 0 = ocean)
 * estep - determines steepness of equilibrium theta-e profile
 * dorad - do radiation flag
 * frad - flag indicating fixed radiation
 * radcool - fixed radiation cooling rate
 * tpause - height of tropopause
 * radbrk - frac of tpause for which fixed radiative cooling starts decrease
 * cvc - convective mixing rate parameter
 * cvs - stratiform rain rate parameter
 * cve - rain evaporation rate parameter
 * srh - relative humidity at which rain starts
 * wtrain - terminal velocity of raindrops
 * wtsnow - terminal velocity of snowflakes
 * theslop - range of theta_e going from suppressed to full deep convection
 * pscale - height at which precip begins
 * pbltop - top of planetary boundary layer
 * cdrag - surface drag coefficient, used also for thermodynamics
 * wscale - gustiness parameter for surface fluxes
 * cld - cloud absorption parameter
 * cfract - cloud fractional area coverage
 * dt - time step
 * zc - cell center z values
 * xcell - vertical cell sizes in calling model space
 * th - potential temperature profile - cell centered, as all that follow
 * rt - total water mixing ratio profile
 * pi - exner function profile
 * u - x wind profile
 * v - y wind profile
 * xrho - mass density profile in calling model space
 * sth - theta source term profile, return variable as all below
 * sthe_cu - theta-e source term profile from convection
 * sthe_rad - theta-e source term profile from radiation
 * srt - total cloud water mixing ratio source term profile
 * su - x momentum source term profile
 * sv - y momentum source term profile
 * irflux - radiative flux profile
 * rain - rainfall rate
 * eflux - surface theta-e flux
 * rflux - surface total water flux (evaporation)
 * uflux - surface x momentum flux
 * vflux - surface y momentum flux
 * convthrot - convective throttle value
 * fluxthrot - flux throttle value
 * cvcaug - deep convective mixing rate
 * stemp - surface temperature (K)
 * swater - surface water column (land) (mm = km/m^2)
 */

/* constants for land surface model */
#define WCRIT 10.         /* mm or kg/m^2 of standing water */
#define SB 5.67e-8        /* stefan-boltzmann constant */

void diabat3(TLONG nz, TLONG land, TFLOAT estep, TLONG dorad, TLONG frad,
	     TFLOAT radcool, TFLOAT tpause, TFLOAT radbrk, TFLOAT cvc,
	     TFLOAT cvs, TFLOAT cve, TFLOAT srh, TFLOAT wtrain,
	     TFLOAT wtsnow, TFLOAT theslop, TFLOAT pscale,
             TFLOAT pbltop, TFLOAT cdrag, TFLOAT wscale, TFLOAT cld,
             TFLOAT cfract, TFLOAT lamtemp, TFLOAT lamsea, TFLOAT lamwater,
	     TFLOAT albedo, TFLOAT dt, TFLOAT *zc, TFLOAT *xcell,
	     TFLOAT *th, TFLOAT *rt, TFLOAT *pi, TFLOAT *u, TFLOAT *v,
	     TFLOAT *xrho, TFLOAT *sth, TFLOAT *sthe_cu, TFLOAT *sthe_rad,
	     TFLOAT *srt, TFLOAT *su, TFLOAT *sv, TFLOAT *irflux,
	     TFLOAT *rain, TFLOAT *srain, TFLOAT *erain,
	     TFLOAT *eflux, TFLOAT *rflux, TFLOAT *uflux, TFLOAT *vflux,
	     TFLOAT *convthrot, TFLOAT *fluxthrot, TFLOAT *cvcaug,
	     TFLOAT *ctopdeep, TFLOAT *ctopshallow,
	     TFLOAT *stemp, TFLOAT *swater)
{
  static TLONG iz;            /* index variables */
  static TFLOAT *the;         /* equivalent potential temperature */
  static TFLOAT *rsat;        /* saturation mixing ratio */
  static TFLOAT *sthe_radclr; /* clear sky radiation source */
  static TFLOAT *irfluxclr;   /* clear sky radiation flux */
  static TFLOAT *dz;          /* vertical cell size in geometric space */
  static TFLOAT *rho;         /* geometric space density */
  static TFLOAT surfpi, surfth;  /* surface stuff for radiation */
  static TFLOAT landrh;       /* land relative humidity */
  static TFLOAT irref;         /* reference IR radiation value */
  static TFLOAT thsane;       /* limit range of potential temp for thermo */
  static int begin = 1;

  /* make space only once */
  if (begin) {
    the = floatbuff(nz);
    rsat = floatbuff(nz);
    sthe_radclr = floatbuff(nz);
    irfluxclr = floatbuff(nz);
    dz = floatbuff(nz);
    rho = floatbuff(nz);

    /* radiation routines need to be called on first time step */
    if (dorad == 0) {
      fprintf(stderr,"diabat3: dorad must be nonzero on first call\n");
      exit(1);
    }
    begin = 0;
  }

  /* compute the land relative humidity -- 1 over ocean */
  if (land == 1) {
    landrh = 1. - exp(- *swater/WCRIT);
  }
  else {
    landrh = 1.;
  }

  /* compute the equivalent potential temperature and
     saturation mixing ratio */
  for (iz = 0; iz < nz; iz++) {
    thsane = th[iz];
    if (thsane < 230.) thsane = 230.;
    if (thsane > 400.) thsane = 400.;
    rsat[iz] = rs(thsane,pi[iz]);
    the[iz] = thetaecomp2(th[iz],rt[iz],rsat[iz]);

    /* compute cell sizes and geometric density */
    rho[iz] = PREF*pow(pi[iz]/CP,1./KAPPA)/(KAPPA*th[iz]*pi[iz]);
    dz[iz] = xrho[iz]*xcell[iz]/rho[iz];
  }

  /* call the convective parameterization */
  cuparam(nz,estep,cvc,cvs,cve,srh,wtrain,wtsnow,theslop,pscale,
	  pbltop,cdrag,wscale,*stemp,landrh,dt,zc,dz,th,the,rt,pi,u,v,
	  rho,rsat,sthe_cu,srt,su,sv,rain,srain,erain,eflux,rflux,
	  uflux,vflux,convthrot,fluxthrot,cvcaug,ctopdeep,ctopshallow);

  /* do radiation only if specified */
  if (dorad) {
    
    /* use fixed radiation if frad not zero */
    if (frad) {
      fradiation(nz,zc,radcool,tpause,radbrk,sthe_rad,sthe_radclr,
		 irflux,irfluxclr);
    }
    
    /* otherwise call the radiative parameterization */
    else {
      
      /* cloudy atmosphere radiation (only compute when there are clouds) */
      if (cfract > 0.) {
	radiation(nz,0,cld,*stemp,albedo,zc,dz,th,the,rt,pi,rho,rsat,
		  sthe_rad,irflux);
      }
      
      /* clear atmosphere radiation (only compute when there is clear air) */
      if (cfract < 1.) {
	radiation(nz,0,0.,*stemp,albedo,zc,dz,th,the,rt,pi,rho,rsat,
		  sthe_radclr,irfluxclr);
      }
    }
    
    /* the actual radiation is a weighted average of cloudy and clear
       (don't do anything if cfract = 1) */
    if (cfract == 0.) {
      for (iz = 0; iz < nz; iz++) {
	sthe_rad[iz] = sthe_radclr[iz];
	irflux[iz] = irfluxclr[iz];
      }
    }
    else if (cfract < 1.) {
      for (iz = 0; iz < nz; iz++) {
	sthe_rad[iz] = cfract*sthe_rad[iz] + (1. - cfract)*sthe_radclr[iz];
	irflux[iz] = cfract*irflux[iz] + (1. - cfract)*irfluxclr[iz];
      }
    }
  }
  
  /* compute the new surface temperature and water column over land --
   * over sea a new surface temperature is calculated but the 
   * water column is not since it doesn't mean anything there */
  irref = SB*(*stemp)*(*stemp)*(*stemp)*(*stemp);
  if (land) {
    *stemp += - dt*lamtemp*(*stemp)*(irflux[0] + pi[0]*(*eflux))/irref;
    *swater += dt*(1000.*((*rain) - (*rflux)) - lamwater*(*swater));
  }
  else {
    *stemp += - dt*lamsea*(*stemp)*(irflux[0] + pi[0]*(*eflux))/irref;
  }

  /* compute the theta source term -- note that the formula below depends
   * on the definition of theta-e used in this program --
   * compute sth every time since sthe_cu and srt are updated every time
   */
  for (iz = 0; iz < nz; iz++) {
    sth[iz] = th[iz]*((sthe_cu[iz] + sthe_rad[iz])/the[iz] - ALFA*srt[iz]);
  }
}
