/* getprofile -- Get the possibly time-dependent reference profile
 * from a file.  If it depends on time, then the reference profile
 * gets updated continuously.  Otherwise, it is just an initialization.
 * Copied from conv and doubles converted to doubles.
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <quickcandis.h>

static int uninitialized = 1;
static double delt;
static double *thin, *rtin, *vxin, *vyin;
static long nz, ntime;

double *doublebuff(long jsize);
void fbuf(double *ptr);

static double *fillfield(int handle, long nz, long ntime,
			 int nfields, char *fname);

void getprofile(double time, double *th0, double *rt0,
		double *vx0, double *vy0)
{
  long iz, it;

  /* check for initialization */
  if (uninitialized) {
    fprintf(stderr,"conv: oops! call to getprofile before readprofile\n");
    exit(1);
  }
  
  /* time independent case */
  if (ntime == 0) {
    for (iz = 0; iz < nz; iz++) {
      th0[iz] = thin[iz];
      rt0[iz] = rtin[iz];
      vx0[iz] = vxin[iz];
      vy0[iz] = vyin[iz];
    }
  }

  /* time dependent case */
  else {
    it = time/delt + 0.5;
    if (it < 0) {
      fprintf(stderr,"conv: oops, a negative time!\n");
      exit(1);
    }
    
    /* time is mod the max time in the initialization input */
    while (it >= ntime) it -= ntime;

    /* since the time is the 2nd index field, it iterates most rapidly */
    for (iz = 0; iz < nz; iz++) {
      th0[iz] = thin[iz*ntime + it];
      rt0[iz] = rtin[iz*ntime + it];
      vx0[iz] = vxin[iz*ntime + it];
      vy0[iz] = vyin[iz*ntime + it];
    }
  }
}

void readprofile(long nz1, char *profname)
{
  int handle, ndims, nfields;
  long it;
  float *tptr;

  /* open the profile Candis file */
  handle = candisread(profname);

  /* check the number of dimensions */
  ndims = candisdinfo(handle);
  if (ndims < 1 || ndims > 2) {
    fprintf(stderr,"conv: number of profile file dimensions not 1 or 2\n");
    exit(1);
  }

  /* check z dimension name and size */
  nz = nz1;
  if (strcmp(candisdnameget(handle,0), "z") != 0) {
    fprintf(stderr,"conv: profile file z dimension name not correct\n");
    exit(1);
  }
  if (candisdsizeget(handle,0) != nz) {
    fprintf(stderr,"conv: profile file z dimension size not correct\n");
    exit(1);
  }

  /* check time dimension name and get size if it exists */
  if (ndims == 2) {
    if (strcmp(candisdnameget(handle,1), "time") != 0) {
      fprintf(stderr,"conv: profile file time dimension name not correct\n");
      exit(1);
    }

    /* get the time index field values and size -- assumes equally spaced
     * times
     */
    ntime = candisdsizeget(handle, 1);
    tptr = candisdvalsget(handle, 1);
    delt = (tptr[ntime - 1] - tptr[0])/ntime;
  }
  else ntime = 0;

  /* get the number of fields */
  nfields = candisfinfo(handle);

  /* fill each desired field */
  thin = fillfield(handle, nz, ntime, nfields, "th");
  rtin = fillfield(handle, nz, ntime, nfields, "rt");
  vxin = fillfield(handle, nz, ntime, nfields, "vx");
  vyin = fillfield(handle, nz, ntime, nfields, "vy");

  /* we're done */
  uninitialized = 0;
  return;
}

/* fillfield -- Get the data from a specified field */
static double *fillfield(int handle, long nz, long ntime,
			 int nfields, char *fname)
{
  int i, j;
  long mtime;
  float *fpointer;
  double *buffer;

  /* make buffer */
  mtime = ntime > 0 ? ntime : 1;
  buffer = doublebuff(mtime*nz);

  /* see if there is a field by the desired name */
  for (i = 0; i < nfields; i++) {
    if (strcmp(candisfnameget(handle, i), fname) == 0) {

      /* does it have the right size? */
      if (candisfsizeget(handle, i) != mtime*nz) {
	fprintf(stderr,"conv: initial profilefield %s has wrong size\n",
		fname);
	fprintf(stderr,"%d %d\n",candisfsizeget(handle, i),mtime*nz);
	exit(1);
      }

      /* transfer the values */
      fpointer = candisfvalsget(handle, i);
      for (j = 0; j < mtime*nz; j++) buffer[j] = fpointer[j];

      /* we're done */
      return(buffer);
    }
  }

  /* nope -- fail */
  fprintf(stderr,"conv: no initial profile field named %s\n",fname);
  exit(1);
}
