/* index.c -- Contains bounds checking index functions and the hypersmoother.
 */
#include <stdlib.h>
#include <stdio.h>
#include "grid.h"
#include "index.h"

static int nx, ny, nth;

int iii(int nx, int ny, int nth, int ix, int iy, int ith)
{
  if (ix < -2 || iy < -2 || ith < -1 ||
      ix > nx + 1 || iy > ny + 1 || ith > nth) {
    abort();
  }
  else return ((ith) + 1 + (nth + 4)*((iy) + 2 + (ny + 4)*((ix) + 2)));
}

int jjj(int nx, int ny, int nth, int ix, int iy, int ith)
{
  if (ix < 0 || iy < 0 || ith < 0 || ix > nx || iy > ny || ith > nth) {
    abort();
  }
  else return ((ith) + (nth + 1)*((iy) + (ny + 1)*(ix)));
}

int ii(int nx, int ny, int ix, int iy)
{
  if (ix < -2 || iy < -2 || ix > nx + 1 || iy > ny + 1) {
    abort();
  }
  else return ((iy) + 2 + (ny + 4)*((ix) + 2));
}

int jj(int nx, int ny, int ix, int iy)
{
  if (ix < 0 || iy < 0 || ix > nx || iy > ny) {
    abort();
  }
  else return ((iy) + (ny + 1)*(ix));
}

int fiy(int ny, int iy)
{
  if (iy < -2 || iy > ny + 1) abort();
  else return ((iy) + 1);
}

int fjy(int ny, int iy)
{
  if (iy < 0 || iy > ny) abort();
  else return (iy);
}

int fk(int ny, int nth, int iy, int ith)
{
  if (iy < 0 || iy >= ny || ith < 0 || ith >= nth) abort();
  else return((ith) + nth*(iy));
}

int fk2(int nx, int ny, int ix, int iy)
{
  if (ix < 0 || ix >= nx || iy < 0 || iy >= ny) abort();
  else return((iy) + ny*(ix));
}





/* vsmooth -- do conservative vertical hypersmoothing */
void vsmooth(int nth, double *fld, double lambda)
{
  int ith, im1, im2, ip1, ip2;
  double lambda1;
  static double *fld1;
  static begin = 1;
  
  /* only allocate memory once */
  if (begin) {
    fld1 = grid_alloc(nth);
    begin = 0;
  }

  lambda1 = 1. - lambda;

  /* do smoothing */
  for (ith = 0; ith < nth; ith++) {
    if (ith == nth - 1) {
      im1 = ith - 1;
      im2 = ith - 2;
      ip1 = ith;
      ip2 = ith - 1;
    }
    else if (ith == nth - 2) {
      im1 = ith - 1;
      im2 = ith - 2;
      ip1 = ith + 1;
      ip2 = ith + 1;
    }
    else if (ith == 0) {
      im1 = ith;
      im2 = ith + 1;
      ip1 = ith + 1;
      ip2 = ith + 2;
    }
    else if (ith == 1) {
      im1 = ith - 1;
      im2 = ith - 1;
      ip1 = ith + 1;
      ip2 = ith + 2;
    }
    else {
      im1 = ith - 1;
      im2 = ith - 2;
      ip1 = ith + 1;
      ip2 = ith + 2;
    }
    fld1[ith] = lambda1*fld[ith] + lambda*(-(fld[ip2] + fld[im2])/6. +
					   2.*(fld[ip1] + fld[im1])/3.);
  }

  for (ith = 0; ith < nth; ith++) {
    fld[ith] = fld1[ith];
  }

#if 0
  grid_free(fld1);
#endif
}

/* hsmooth -- do horizontal conservative hypersmoothing */
void hsmooth(int nx, int ny, int nth, int channel, double *fld, double kappa)
{
  int ix, iy, ith;
  double kappa1;
  static double *tfld;
  static begin = 1;
  
  /* only allocate memory once */
  if (begin) {
    tfld = grid_alloc(nx*ny);
    begin = 0;
  }

  /* complement to kappa */
  kappa1 = 1. - kappa;

  /* loop on height */
  for (ith = 0; ith < nth; ith++) {

    /* hypersmooth on x and y */
    for (ix = 0; ix < nx; ix++) {
      for (iy = 0; iy < ny; iy++) {
	tfld[K2(ix,iy)] = kappa1*fld[I(ix,iy,ith)] -
	  kappa*((fld[I(ix + 2,iy,ith)] + fld[I(ix - 2,iy,ith)] +
		  fld[I(ix,iy + 2,ith)] + fld[I(ix,iy - 2,ith)]) +
		  2.*(fld[I(ix + 1,iy + 1,ith)] + fld[I(ix + 1,iy - 1,ith)] +
		      fld[I(ix - 1,iy + 1,ith)] + fld[I(ix - 1,iy - 1,ith)]) -
		  8.*(fld[I(ix + 1,iy,ith)] + fld[I(ix - 1,iy,ith)] +
		      fld[I(ix,iy + 1,ith)] + fld[I(ix,iy - 1,ith)]))/20.;
      }
    }

    /* copy from temp field back to original location */
    for (iy = 0; iy < ny; iy++) {
      for (ix = 0; ix < nx; ix++) {
	fld[I(ix,iy,ith)] = tfld[K2(ix,iy)];
      }
    }
  }

  /* free space */
#if 0
  grid_free(tfld);
#endif
}
