/* init.c -- Initialize fields.
 */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <quickcandis.h>
#include "params.h"
#include "vert.h"
#include "thermo.h"
#include "index.h"
#include "init.h"

/* limiting values on vertical potential temp gradient (K/km) */
#define THGRADMIN 1.0
#define THGRADMAX 20.0

/* scale height for RH random initialization */
#define RHSCALEHT 3.0

/* parameters */
static int nx, ny, nth, id1, channel;
static double xbase, ybase, xctr, yctr, thb0, tht0, dx, dy;
static double srftmp0, srfmag, srfwid, srfoff;
static double hmax, thtp0, psurf, ptp, ptop, rhinit;
static double rhrand, etarand, dsrftmp;
static double vx0, vx1, vx2, vxsin;
static double qscale;

/* compute a positive cosine form */
static double cosform(double x)
{
  if (fabs(x) < 1.) return((1. + cos(3.14159*x))/2.);
  else return(0.);
}

/* make a smooth transition between lo value and hi value over the
 * range in tvalue [-slop, +slop]
 */
static double switchit(double tvalue, double slop, double lo, double hi)
{
  double eps;
  eps = tvalue/(2.*slop) + 0.5;
  if (eps > 1.) eps = 1.;
  if (eps < 0.) eps = 0.;
  return(eps*hi + (1. - eps)*lo);
}

/* random number generator uniform on -1 to 1 */
static double myrand(void)
{
  double rvalue;
  rvalue = 2.*(((double)random())/((double)RAND_MAX)) - 1.;
  return(rvalue);
}

/* initialize fields */
void init_fields(double *thb, double *srftmp, double *landfrac,
		 double *topo, double *swater, double *eta,
		 double *vx,  double *vy, double *ssig, double *pi,
		 double *mont, double *z, double *th, double *the,
		 double *thes, double *rt, double *rsat, double *ux,
		 double *uy, double *ert, double *etamin, double *etamax,
		 double *vxb, double *vyb, double *rain, double *srain,
		 double *prain, double *erain, double *rain1,
		 double *rain2s, double *rain2t, double *eflux, double *rflux,
		 double *vxflux, double *vyflux, double *convthrot,
		 double *fluxthrot, double *cvcaug, double *ctopdeep,
		 double *ctopshallow)
{
  int ix, iy, ith, i, i2, k, k2;
  float fiy;
  int deftrop, defcol, defzone, defsst, landflg;
  int sfilehandle, nvars, loop;
  char sfile[LINE];
  float *etafld, *rtfld, *vxfld, *vyfld, *ssigfld, *thbfld;
  float *topofld, *srftmpfld, *lsmaskfld;
  double xc, yc, rcsq, rc, sigma, xrad, yrad, etas, etat, dsig, tpsig;
  double pimean, count, rhomean, thmean, yrel, znondim;

  /* get the parameters */
  nx = params_int("nx");
  ny = params_int("ny");
  nth = params_int("nth");
  xbase = params_float("xbase");
  ybase = params_float("ybase");
  xctr = params_float("xctr");
  yctr = params_float("yctr");
  tht0 = params_float("tht0");
  thb0 = params_float("thb0");
  psurf = params_float("psurf");
  ptop = params_float("ptop");
  dx = params_float("dx");
  dy = params_float("dy");
  hmax = params_float("hmax");
  landflg = params_int("landflg");
  xrad = params_float("xrad");
  yrad = params_float("yrad");
  vx0 = params_float("vx0");
  vx1 = params_float("vx1");
  vx2 = params_float("vx2");
  vxsin = params_float("vxsin");
  channel = params_int("channel");
  srftmp0 = params_float("srftmp0");
  srfmag = params_float("srfmag");
  srfwid = params_float("srfwid");
  srfoff = params_float("srfoff");
  rhrand = params_float("rhrand");
  etarand = params_float("etarand");
  dsrftmp = params_float("dsrftmp");
  deftrop = params_int("deftrop");
  defcol = params_int("defcol");
  defzone = params_int("defzone");
  defsst = params_int("defsst");
  id1 = params_int("id1");

  /* select initialization method */
  if (deftrop + defcol + defzone != 1) {
    fprintf(stderr,"init: choose one and only one initialization method\n");
    exit(1);
  }

  /* id1 is not consistent with defcol */
  if (id1 && defcol) {
    fprintf(stderr,"init: id1 is not consistent with defcol\n");
    exit(1);
  }

  /* simplest method -- fix tropopause, lapse rate, and RH */
  if (deftrop) {

    /* extract parameters defining tropopause and initial relative humidity */
    thtp0 = params_float("thtp0");
    ptp = params_float("ptp");
    rhinit = params_float("rhinit");

    /* extract qscale for the ideal1 parameterization initialization */
    if (id1) qscale = params_float("qscale");
  }

  /* fix the density and mixing ratio in a column */
  else if (defcol) {

    /* extract data about eta and ert profiles from a candis file */
    strncpy(sfile,params_string("cfile"),LINE - 1);
    sfile[LINE - 1] ='\0';
    sfilehandle = candisread(sfile);
    if (candisdinfo(sfilehandle) != 1) {
      fprintf(stderr,"init: file %s needs to have 1 dimension\n",sfile);
      exit(1);
    }
    if (strcmp(candisdnameget(sfilehandle,0),"sigma") != 0) {
      fprintf(stderr,"init: dimension in file %s must have name 'sigma'\n",
	     sfile);
      exit(1);
    }
    if (candisdsizeget(sfilehandle,0) != nth) {
      fprintf(stderr,"init: dimension size wrong in file %s\n",sfile);
      exit(1);
    }
    nvars = candisfinfo(sfilehandle);
    etafld = rtfld = (float*)NULL;
    for (loop = 0; loop < nvars; loop++) {
      if (strcmp(candisfnameget(sfilehandle,loop),"eta") == 0) {
	etafld = candisfvalsget(sfilehandle,loop);
      }
      if (strcmp(candisfnameget(sfilehandle,loop),"rt") == 0) {
	rtfld = candisfvalsget(sfilehandle,loop);
      }
      if (strcmp(candisfnameget(sfilehandle,loop),"thb") == 0) {
	thbfld = candisfvalsget(sfilehandle,loop);
      }
    }
    if (etafld == (float*)NULL || rtfld == (float*)NULL ||
	thbfld == (float*)NULL) {
      fprintf(stderr,"init: oops, no 'eta', 'rt', or 'thb' field\n");
      exit(1);
    }
  }

  /* fix zonally symmetric eta, rt, vx, vy, ssig, thb from a candis file */
  else if (defzone) {
    strncpy(sfile,params_string("zfile"),LINE - 1);
    sfile[LINE - 1] ='\0';
    sfilehandle = candisread(sfile);
    if (candisdinfo(sfilehandle) != 2) {
      fprintf(stderr,"init: file %s needs to have 2 dimensions\n",sfile);
      exit(1);
    }
    if (strcmp(candisdnameget(sfilehandle,0),"y") != 0) {
      fprintf(stderr,"init: dimension 0 in file %s must have name 'y'\n",
	     sfile);
      exit(1);
    }
    if (strcmp(candisdnameget(sfilehandle,1),"sigma") != 0) {
      fprintf(stderr,"init: dimension 1 in file %s must have name 'sigma'\n",
	     sfile);
      exit(1);
    }
    if (candisdsizeget(sfilehandle,0) != ny) {
      fprintf(stderr,"init: dimension 0 size wrong in file %s\n",sfile);
      exit(1);
    }
    if (candisdsizeget(sfilehandle,1) != nth) {
      fprintf(stderr,"init: dimension 1 size wrong in file %s\n",sfile);
      exit(1);
    }
    nvars = candisfinfo(sfilehandle);
    etafld = rtfld = vxfld = vyfld = ssigfld = thbfld = (float*)NULL;
    for (loop = 0; loop < nvars; loop++) {
      if (strcmp(candisfnameget(sfilehandle,loop),"eta") == 0) {
	etafld = candisfvalsget(sfilehandle,loop);
      }
      if (strcmp(candisfnameget(sfilehandle,loop),"rt") == 0) {
	rtfld = candisfvalsget(sfilehandle,loop);
      }
      if (strcmp(candisfnameget(sfilehandle,loop),"vx") == 0) {
	vxfld = candisfvalsget(sfilehandle,loop);
      }
      if (strcmp(candisfnameget(sfilehandle,loop),"vy") == 0) {
	vyfld = candisfvalsget(sfilehandle,loop);
      }
      if (strcmp(candisfnameget(sfilehandle,loop),"ssig") == 0) {
	ssigfld = candisfvalsget(sfilehandle,loop);
      }
      if (strcmp(candisfnameget(sfilehandle,loop),"thb") == 0) {
	thbfld = candisfvalsget(sfilehandle,loop);
      }
    }

    if (etafld == (float*)NULL || rtfld == (float*)NULL ||
	vxfld == (float*)NULL || vyfld == (float*)NULL ||
	ssigfld == (float*)NULL || thbfld == (float*)NULL) {
      fprintf(stderr,"init: missing a zonally symmetric input field\n");
      exit(1);
    }
  }

  /* you shouldn't get here! */
  else {
    fprintf(stderr,"init: need to define some initialization method\n");
    exit(1);
  }

  /* compute aux vars */
  dsig = 1./nth;
  tpsig = (thtp0 - thb0)/(tht0 - thb0);

  /* define dependent fields */
  for (ix = 0; ix < nx; ix++) {
    for (iy = 0; iy < ny; iy++) {
      i2 = I2(ix,iy);

      /* zero fluxes */
      rain[i2] = srain[i2] = prain[i2] = erain[i2] =
	eflux[i2] = rflux[i2] = vxflux[i2] = vyflux[i2] = 0.;
      rain1[i2] = rain2s[i2] = rain2t[i2] = convthrot[i2] =
	fluxthrot[i2] = cvcaug[i2] = ctopdeep[i2] = ctopshallow[i2] = 0.;

      /* vertical grid -- defined tropopause */
      if (deftrop) {
	thb[i2] = thb0;
	etas = (tht0 - thb[i2])*(ptp - ptop)/(GEE*(tht0 - thtp0));
	etat = (tht0 - thb[i2])*(psurf - ptp)/(GEE*(thtp0 - thb[i2]));
	for (ith = 0; ith < nth; ith++) {
	  i = I(ix,iy,ith);

	  /* eta field */
	  eta[i] = switchit(SIGC(ith) - tpsig,2.*dsig,etat,etas);
	}
      }

      /* eta and ert profiles as well as thb are externally specified */
      if (defcol) {
	thb[I2(ix,iy)] = thbfld[0];
	for (ith = 0; ith < nth; ith++) {
	  i = I(ix,iy,ith);
	  eta[i] = etafld[ith];
	}
      }

      /* eta, ert, vx, vy, ssig, and thb externally specified */
      if (defzone) {
	thb[I2(ix,iy)] = thbfld[iy];
	for (ith = 0; ith < nth; ith++) {
	  i = I(ix,iy,ith);
	  k = K(iy,ith);
	  eta[i] = etafld[k];
	}
      }

      /* exner function, montgomery potential, and z on centered grid */
      i = I(ix,iy,0);
      vert_icomp(thb[i2],topo[i2],&eta[i],&pi[i],&mont[i],&z[i]);

      /* vertical grid again */
      for (ith = 0; ith < nth; ith++) {
	i = I(ix,iy,ith);
	k = K(iy,ith);
	sigma = SIGC(ith);

	/* vx, vy, and ssig fields */
	if (channel) {
	  fiy = iy + 0.5;
	}
	else {
	  fiy = iy;
	}
	vx[i] = vx0 + vx1*sigma + vx2*sigma*sigma/2. +
	  vxsin*sin(6.28318*fiy/ny);
	vy[i] = ssig[i] = 0.;

	/* potential temperature and total cloud water mixing ratio field */
	th[i] = thb[i2] + SIGC(ith)*(tht0 - thb[i2]);
	rsat[i] = rs(th[i],pi[i]);

	/* defined tropopause -- different profile if ideal1 param used */
	if (deftrop) {
	  if (id1) {
	    rt[i] = rhinit*rsat[I(ix,iy,0)]*exp(-z[i]/qscale);
	    if (rt[i] > rsat[i]) rt[i] = rsat[i];
	  }
	  else {
	    rt[i] = rhinit*rsat[i];
	  }
	}

	/* defined profiles */
	if (defcol) {
	  rt[i] = rtfld[ith];
	}

	/* defined zonal pattern */
	if (defzone) {
	  rt[i] = rtfld[k];
	  vx[i] = vxfld[k];    /* these 3 overwrite earlier specifications */
	  vy[i] = vyfld[k];
	  ssig[i] = ssigfld[k];
	}

	/* random initialization of humidity and density */
	znondim = z[i]/RHSCALEHT;
	rt[i] *= (1. + rhrand*myrand()*znondim*exp(1. - znondim));
	eta[i] *= (1. + etarand*myrand()*znondim*exp(1. - znondim));

	/* compute initial equivalent potential temperatures */
	the[i] = thetaecomp(th[i],rt[i],pi[i]);
	thes[i] = thetaecomp(th[i],rsat[i],pi[i]);

	/* create flux variables */
	ux[i] = eta[i]*vx[i];
	uy[i] = eta[i]*vy[i];
	ert[i] = eta[i]*rt[i];
      }

      /* initialize vxb, vyb */
      vxb[i2] = SURFEXTRAP(vx,ix,iy);
      vyb[i2] = SURFEXTRAP(vy,ix,iy);
    }
  }

  /* initialize topography and surface temperature */
  if (defsst) {

    /* get surface temperature and topography from a file */
    strncpy(sfile,params_string("sfile"),LINE - 1);
    sfile[LINE - 1] = '\0';
    sfilehandle = candisread(sfile);
    if (candisdinfo(sfilehandle) != 2) {
      fprintf(stderr,"init: file %s needs to have 2 dimensions\n",sfile);
      exit(1);
    }
    if (strcmp(candisdnameget(sfilehandle,0),"x") != 0) {
      fprintf(stderr,"init: dimension 0 in file %s must have name 'x'\n",
	     sfile);
      exit(1);
    }
    if (strcmp(candisdnameget(sfilehandle,1),"y") != 0) {
      fprintf(stderr,"init: dimension 1 in file %s must have name 'y'\n",
	     sfile);
      exit(1);
    }
    if (candisdsizeget(sfilehandle,0) != nx) {
      fprintf(stderr,"init: dimension 0 size wrong in file %s\n",sfile);
      exit(1);
    }
    if (candisdsizeget(sfilehandle,1) != ny) {
      fprintf(stderr,"init: dimension 1 size wrong in file %s\n",sfile);
      exit(1);
    }
    nvars = candisfinfo(sfilehandle);
    topofld = srftmpfld = lsmaskfld = (float*)NULL;
    for (loop = 0; loop < nvars; loop++) {
      if (strcmp(candisfnameget(sfilehandle,loop),"topo") == 0) {
	topofld = candisfvalsget(sfilehandle,loop);
      }
      if (strcmp(candisfnameget(sfilehandle,loop),"srftmp") == 0) {
	srftmpfld = candisfvalsget(sfilehandle,loop);
      }
      if (strcmp(candisfnameget(sfilehandle,loop),"lsmaskv") == 0) {
	lsmaskfld = candisfvalsget(sfilehandle,loop);
      }
    }
    if (topofld == (float*)NULL || srftmpfld == (float*)NULL ||
	lsmaskfld == (float*)NULL) {
      fprintf(stderr,"init: missing a surface input field\n");
      exit(1);
    }

    /* transfer to the internal model fields */
    for (ix = 0; ix < nx; ix++) {
      for (iy = 0; iy < ny; iy++) {
	i2 = I2(ix,iy);
	k2 = K2(ix,iy);
	topo[i2] = topofld[k2];
	srftmp[i2] = srftmpfld[k2];
	swater[i2] = 0.;
	landfrac[i2] = 1. - lsmaskfld[k2];
      }
    }    
  }
  else {

    /* do a simple initialization */
    for (ix = 0; ix < nx; ix++) {
      for (iy = 0; iy < ny; iy++) {
	i2 = I2(ix,iy);
	xc = (xbase + dx*ix - xctr)/xrad;
	yc = (ybase + dy*iy - yctr)/yrad;
	rcsq = xc*xc + yc*yc;
	rc = pow(rcsq,0.5);
	
	/* topography */
	topo[i2] = hmax*exp(-rcsq);
	
	/* surface temperature */
	yrel = (iy*dy + ybase - srfoff)/srfwid;
	srftmp[i2] = srftmp0 +
	  srfmag*(1.3333*cosform(yrel) - 0.3333*cosform(2.*yrel)) +
	  dsrftmp*cosform(rc);

	/* surface moisture storage */
	swater[i2] = 0.;

	/* land fraction */
	if (landflg && rcsq < 1.) {
	  landfrac[i2] = 1.;
	}
	else {
	  landfrac[i2] = 0.;
	}
      }
    }
  }

  /* compute limiting profiles for eta */
  for (ith = 0; ith < nth; ith++) {
    pimean = 0.;
    count = 0.;
    for (ix = 0; ix < nx; ix++) {
      for (iy = 0; iy < ny; iy++) {
	pimean += pi[I(ix,iy,ith)];
	count += 1.;
      }
    }
    pimean /= count;
    thmean = SIGC(ith)*(tht0 - thb0) + thb0;
    rhomean = PREF*pow(pimean/CP,1./KAPPA)/(KAPPA*thmean*pimean);
    etamin[ith] = rhomean*(tht0 - thb0)/THGRADMAX;
    etamax[ith] = rhomean*(tht0 - thb0)/THGRADMIN;
  }
}
