/* source.c -- Compute source terms.
 */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "params.h"
#include "diabat3.h"
#include "ideal1.h"
#include "vert.h"
#include "thermo.h"
#include "source.h"
#include "grid.h"
#include "index.h"
#include "wallclock.h"

#define RADITERMAX 5

static int nx, ny, nth, frad;
static double xbase, ybase, dx, dy, dsig, dt;
static double sthmax, hmax, tmax, xrad, yrad, sigrad;
static double xctr, yctr, sigctr, xvel, rcool, tpause, tbreak;
static int diab3, id1, raditer;
static double estep, cvc, cvs, cve, srh, wtrain, wtsnow, theslop;
static double pscale, pbltop, cld, cfract, frict, cdrag, ctherm, wscale;
static double mrelax, trop, skew, qscale, d, mucin, epsilon;
static double lamtemp, lamsea, lamwater, albedo, lambdas, delq, vx0;
static double mu, lam;

/* local smoothing program */
static void smooth(int nth, double *fld)
{
  static double *fld1;
  int ith;
  static int begin = 1;

  /* allocate space only once */
  if (begin) {
    fld1 = grid_alloc(nth);
    begin = 0;
  }

  /* do smoothing except end points (this is not quite conservative) */
  for (ith = 1; ith < nth - 1; ith++) {
    fld1[ith] = 0.5*fld[ith] + 0.25*(fld[ith + 1] + fld[ith - 1]);
  }

  /* copy back */
  for (ith = 1; ith < nth - 1; ith++) {
    fld[ith] = fld1[ith];
  }

  /* free space */
#if 0
  grid_free(fld1);
#endif
}



/* initialize source term calculator */
void source_init()
{
  nx = params_int("nx");
  ny = params_int("ny");
  nth = params_int("nth");
  xbase = params_float("xbase");
  ybase = params_float("ybase");
  dx = params_float("dx");
  dy = params_float("dy");
  dt = params_float("dt");
  sthmax = params_float("sthmax");
  hmax = params_float("hmax");
  xrad = params_float("xrad");
  yrad = params_float("yrad");
  sigrad = params_float("sigrad");
  xctr = params_float("xctr");
  yctr = params_float("yctr");
  sigctr = params_float("sigctr");
  xvel = params_float("xvel");
  tmax = params_float("tmax");
  diab3 = params_int("diab3");
  id1 = params_int("id1");

  /* at most one parameterization should be activated */
  id1 = id1 == 0 ? 0 : 1;
  diab3 = diab3 == 0 ? 0 : 1;
  if (diab3 + id1 > 1) {
    fprintf(stderr,"source: only one parameterization should be active\n");
    exit(1);
  }

  /* particular parameterizations */
  if (diab3) {
    estep = params_float("estep");
    cvc = params_float("cvc");
    cvs = params_float("cvs");
    cve = params_float("cve");
    srh = params_float("srh");
    wtrain = params_float("wtrain");
    wtsnow = params_float("wtsnow");
    theslop = params_float("theslop");
    pscale = params_float("pscale");
    pbltop = params_float("pbltop");
    cdrag = params_float("cdrag");
    wscale = params_float("wscale");
    cld = params_float("cld");
    cfract = params_float("cfract");
    lamtemp = params_float("lamtemp");
    lamsea = params_float("lamsea");
    lamwater = params_float("lamwater");
    albedo = params_float("albedo");
    frict = params_float("frict");
  }
  if (id1) {
    wscale = params_float("wscale");
    ctherm = params_float("ctherm");
    mrelax = params_float("mrelax");
    trop = params_float("trop");
    skew = params_float("skew");
    qscale = params_float("qscale");
    d = params_float("d");
    mucin = params_float("mucin");
    epsilon = params_float("epsilon");
    lambdas = params_float("lambdas");
    delq = params_float("delq");
    vx0 = params_float("vx0");
    mu = params_float("mu");
    lam = params_float("lam");
  }
  frad = params_int("frad");
  if (frad) {
    rcool = params_float("rcool");
    tpause = params_float("tpause");
    tbreak = params_float("tbreak");
  }

  /* auxiliary calculations */
  dsig = 1./nth;
  raditer = RADITERMAX;
}





/* compute the source terms in the specified column */
void source_comp(double time, double *thb, double *srftmp,
		 double *landfrac, double *topo, double *swater,
		 double *eta, double *pi, double *z, double *mont,
		 double *th, double *the, double *rt, double *vx,
		 double *vy, double *fx, double *fy, double *sth,
		 double *sthe, double *sthe_cu, double *sthe_rad,
		 double *srt, double *rain1, double *rain2s, double *rain2t,
		 double *rain, double *srain, double *erain,
		 double *eflux, double *rflux, double *vxflux,
		 double *vyflux, double *convthrot, double *fluxthrot,
		 double *cvcaug, double *ctopdeep, double *ctopshallow)
{
  int i, im, ip, i2, ix, iy, ith, scount, dorad;
  long land;
  double xc, yc, sigc, x, y, sig, rcsq, smean;
  double thessum, thecnt, pwat, pwatmean, pwatcnt;
  static double *sigcell, *irflux, *seta;
  static double *thesmean;
  static int begin = 1;

  /* allocate space only once */
  if (begin) {
    sigcell = grid_alloc(nth);
    irflux = grid_alloc(nth);
    thesmean = grid_alloc(nth);
    seta = grid_alloc(nth);
    begin = 0;
  }

  /* check if externally specified source term time has expired */
  if (time < tmax) {

    /* compute the initial forcing terms */
    for (ith = 0; ith < nth; ith++) {
      sig = SIGC(ith);
      smean = 0.;
      scount = 0;
      for (ix = 0; ix < nx; ix++) {
	for (iy = 0; iy < ny; iy++) {
	  i = I(ix,iy,ith);
	  x = xbase + dx*ix;
	  y = ybase + dy*iy;
	  xc = (x - xctr - xvel*time)/xrad;
	  yc = (y - yctr)/yrad;
	  sigc = (sig - sigctr)/sigrad;
	  rcsq = xc*xc + yc*yc + sigc*sigc;
	  fx[i] = 0.;
	  fy[i] = 0.;
	  sth[i] = sthmax*exp(-rcsq);
	  smean += sth[i];
	  scount++;
	  sthe[i] = sthe_rad[i] = sthe_cu[i] = 0.;
	  srt[i] = 0.;
	}
      }

      /* make mean of sth equal to zero at each level and zero fluxes */
      smean = smean/scount;
      for (ix = 0; ix < nx; ix++) {
	for (iy = 0; iy < ny; iy++) {
	  i = I(ix,iy,ith);
	  i2 = I2(ix,iy);
	  sth[i] -= smean;
	  rain[i2] = srain[i2] = erain[i2] = 
	    eflux[i2] = rflux[i2] = vxflux[i2] = vyflux[i2] =
	    rain1[i2] = rain2s[i2] = rain2t[i2] = convthrot[i2] =
	    fluxthrot[i2] = cvcaug[i2] = ctopdeep[i2] = ctopshallow[i2] = 0.;
	}
      }
    }
  }

  /* after initial forcing quits, apply a diabatic parameterization
   * add new params by "oring" flag, e. g.,
   * (diab3 || use_something_else) etc.
   */

  /* diabat3 diabatic parameterizations */
  else if (diab3) {

    /* determine whether to call radiation parameterization */
    if (++raditer > RADITERMAX) {
      raditer = 0;
      dorad = 1;
    }
    else {
      dorad = 0;
    }

    /* loop on columns */
    for (ix = 0; ix < nx; ix++) {
      for (iy = 0; iy < ny; iy++) {
	i2 = I2(ix,iy);

	/* compute cell sizes */
	for (ith = 0; ith < nth; ith++) {
	  i = I(ix,iy,ith);
	  sigcell[ith] = dsig;
	  seta[ith] = eta[i];
	}

	/* it is either pure water or pure land at this point */
	if (landfrac[i2] > 0.5) {
	  land = 1;
	}
	else {
	  land = 0;
	}

	/* call parameterization */
	i = I(ix,iy,0);
	diabat3(nth,land,estep,dorad,frad,rcool,tpause,tbreak,cvc,
		cvs,cve,srh,wtrain,wtsnow,theslop,pscale,pbltop,
		cdrag,wscale,cld,cfract,lamtemp,lamsea,lamwater,albedo,dt,
		&z[i],sigcell,&th[i],&rt[i],&pi[i],&vx[i],&vy[i],seta,&sth[i],
		&sthe_cu[i],&sthe_rad[i],&srt[i],&fx[i],&fy[i],irflux,
		&rain[i2],&srain[i2],&erain[i2],&eflux[i2],
		&rflux[i2],&vxflux[i2],&vyflux[i2],&convthrot[i2],
		&fluxthrot[i2],&cvcaug[i2],&ctopdeep[i2],&ctopshallow[i2],
		&srftmp[i2],&swater[i2]);

	/* smooth the vertical profile source terms */
	smooth(nth,&sth[i]);
	smooth(nth,&sthe_cu[i]);
	smooth(nth,&sthe_rad[i]);
	smooth(nth,&srt[i]);
	smooth(nth,&fx[i]);
	smooth(nth,&fy[i]);

	/* combine sthe terms and modulate the friction terms */
	for (ith = 0; ith < nth; ith++) {
	  i = I(ix,iy,ith);
	  sthe[i] = sthe_cu[i] + sthe_rad[i];
	  fx[i] *= frict;
	  fy[i] *= frict;
	}
	vxflux[i2] *= frict;
	vyflux[i2] *= frict;
      }
    }
  }

  /* ideal1 -- implementation of Zeljka's idealized model */
  else if (id1) {

    /* compute mean theta_es */
    for (ith = 0; ith < nth; ith++) {
      thecnt = thessum = 0.;
      for (ix = 0; ix < nx; ix++) {
	for (iy = 0; iy < ny; iy++) {
	  i = I(ix,iy,ith);
	  thecnt += 1.;
	  thessum += thetaescomp(th[i],pi[i]);
	}
      }
      thesmean[ith] = thessum/thecnt;
    }

    /* compute mean precipitable water */
    pwatmean = pwatcnt = 0.;
    for (ix = 0; ix < nx; ix++) {
      for (iy = 0; iy < ny; iy++) {
	pwat = 0.;
	for (ith = 0; ith < nth; ith++) {
	  i = I(ix,iy,ith);
	  pwat += rt[i]*eta[i]*dsig;
	}
	pwatmean += pwat;
	pwatcnt += 1.;
      }
    }
    pwatmean = pwatmean/pwatcnt;

    /* loop on columns */
    for (ix = 0; ix < nx; ix++) {
      for (iy = 0; iy < ny; iy++) {
	i = I(ix,iy,0);
	i2 = I2(ix,iy);
	ideal1(nth,dt,trop,skew,qscale,d,pwatmean,mrelax,mucin,epsilon,lambdas,
	       delq,wscale,ctherm,mu,lam,vx0,&z[i],&th[i],&rt[i],&pi[i],
	       &vx[i],&vy[i],&eta[i],&sth[i],&sthe[i],&srt[i],&fx[i],&fy[i],
	       &rain1[i2],&rain2s[i2],&rain2t[i2],&rain[i2],&eflux[i2],
	       &rflux[i2],&vxflux[i2],&vyflux[i2],thesmean);
      }
    }
  }
	
  /* no parameterization invoked -- zero source terms */
  else {
    for (ix = 0; ix < nx; ix++) {
      for (iy = 0; iy < ny; iy++) {
	for (ith = 0; ith < nth; ith++) {
	  i = I(ix,iy,ith);
	  fx[i] = fy[i] = sth[i] = sthe[i] = srt[i] = 0.;
	  sthe_cu[i] = sthe_rad[i] = 0.;
	}
	i2 = I2(ix,iy);
	rain[i2] = srain[i2] = erain[i2] =
	  eflux[i2] = rflux[i2] = vxflux[i2] = vyflux[i2] =
	  rain1[i2] = rain2s[i2] = rain2t[i2] = convthrot[i2] =
	  fluxthrot[i2] = cvcaug[i2] = ctopdeep[i2] = ctopshallow[i2] = 0.;
      }
    }
  }
}
