/* step.c -- Time stepping module for sigma isentropic model.
 */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <quickcandis.h>
#include "step.h"
#include "params.h"
#include "grid.h"
#include "vert.h"
#include "init.h"
#include "thermo.h"
#include "source.h"
#include "index.h"
#include "wallclock.h"

/* parameters and scalars */
static double xbase, ybase, dx, dy, dt, dh, thb0, tht0;
static double ptop, f0, beta, kappa, spngmag, spngy, spngsig, ckh, ckv, urelax;
static double dsig, lambda;
static int nx, ny, nth;
static int channel;
static int nx1, ny1;

/* constant limiting profiles */
static double *etamin, *etamax;

/* centered 3-D model fields */
static double *vx, *vy, *eta, *the, *thes, *rt, *rsat, *pi, *mont, *z, *th;
static double *ux, *uy, *ert, *kh, *kv;
static double *fssigeta;
static double *fvxux, *fvyux, *fssigux;
static double *fvxuy, *fvyuy, *fssiguy;
static double *fvxert, *fvyert, *fssigert;
static double *fx, *fy, *sth, *sthe, *sthe_cu, *sthe_rad, *srt, *ssig;

/* centered 2-D model fields */
static double *vxb, *vyb, *thb, *thbtemp, *dvx, *dvy;
static double *topo, *sthb, *srftmp, *landfrac, *swater;
static double *rain, *srain, *prain, *erain, *rain1, *rain2s, *rain2t,
  *eflux, *rflux, *vxflux, *vyflux;
static double *convthrot, *fluxthrot, *cvcaug, *ctopdeep, *ctopshallow;
static double *sth1, *srt1, *fx1, *fy1;

/* centered 1-D model fields */
static double *f;

/* temporary vertical profiles */
static double *sth2, *srt2, *fx2, *fy2;

/* temporary zonal mean fields */
static double *vxm, *vxref;

/* corner 3-D fields */
static double *xeta, *xux, *xuy, *xert, *xpi, *xmont, *xz, *xkh, *xkv, *xth;
static double *xfssigeta;
static double *xfvxux, *xfvyux, *xfssigux;
static double *xfvxuy, *xfvyuy, *xfssiguy;
static double *xfvxert, *xfvyert, *xfssigert;
static double *xfvxeta, *xfvyeta;

/* corner 1-D fields */
static double *xf;





/* extend to encompassing grid */
static void c_extend()
{
  int ix, iy, ith, ll, lr, rl, rr, ib, it, ib1, ib2, it1, it2;

  /* extend periodically in x direction */
  for (iy = 0; iy < ny; iy++) {
    for (ith = 0; ith < nth; ith++) {

      /* step 1 */
      ll = I(0,iy,ith);
      lr = I(nx,iy,ith);
      rr = I(nx - 1,iy,ith);
      rl = I(-1,iy,ith);
      eta[lr] = eta[ll];
      eta[rl] = eta[rr];
      ux[lr] = ux[ll];
      ux[rl] = ux[rr];
      uy[lr] = uy[ll];
      uy[rl] = uy[rr];
      fx[lr] = fx[ll];
      fx[rl] = fx[rr];
      fy[lr] = fy[ll];
      fy[rl] = fy[rr];
      sth[lr] = sth[ll];
      sth[rl] = sth[rr];
      sthe[lr] = sthe[ll];
      sthe[rl] = sthe[rr];
      sthe_cu[lr] = sthe_cu[ll];
      sthe_cu[rl] = sthe_cu[rr];
      sthe_rad[lr] = sthe_rad[ll];
      sthe_rad[rl] = sthe_rad[rr];
      srt[lr] = srt[ll];
      srt[rl] = srt[rr];
      ssig[lr] = ssig[ll];
      ssig[rl] = ssig[rr];
      the[lr] = the[ll];
      the[rl] = the[rr];
      thes[lr] = thes[ll];
      thes[rl] = thes[rr];
      ert[lr] = ert[ll];
      ert[rl] = ert[rr];
      rt[lr] = rt[ll];
      rt[rl] = rt[rr];
      rsat[lr] = rsat[ll];
      rsat[rl] = rsat[rr];
      vx[lr] = vx[ll];
      vx[rl] = vx[rr];
      vy[lr] = vy[ll];
      vy[rl] = vy[rr];
      pi[lr] = pi[ll];
      pi[rl] = pi[rr];
      mont[lr] = mont[ll];
      mont[rl] = mont[rr];
      th[lr] = th[ll];
      th[rl] = th[rr];
      z[lr] = z[ll];
      z[rl] = z[rr];
      kh[lr] = kh[ll];
      kh[rl] = kh[rr];
      kv[lr] = kv[ll];
      kv[rl] = kv[rr];

      /* step 2 */
      ll = I(1,iy,ith);
      lr = I(nx + 1,iy,ith);
      rr = I(nx - 2,iy,ith);
      rl = I(-2,iy,ith);
      eta[lr] = eta[ll];
      eta[rl] = eta[rr];
      ux[lr] = ux[ll];
      ux[rl] = ux[rr];
      uy[lr] = uy[ll];
      uy[rl] = uy[rr];
      fx[lr] = fx[ll];
      fx[rl] = fx[rr];
      fy[lr] = fy[ll];
      fy[rl] = fy[rr];
      sth[lr] = sth[ll];
      sth[rl] = sth[rr];
      sthe[lr] = sthe[ll];
      sthe[rl] = sthe[rr];
      sthe_cu[lr] = sthe_cu[ll];
      sthe_cu[rl] = sthe_cu[rr];
      sthe_rad[lr] = sthe_rad[ll];
      sthe_rad[rl] = sthe_rad[rr];
      srt[lr] = srt[ll];
      srt[rl] = srt[rr];
      ssig[lr] = ssig[ll];
      ssig[rl] = ssig[rr];
      the[lr] = the[ll];
      the[rl] = the[rr];
      thes[lr] = thes[ll];
      thes[rl] = thes[rr];
      ert[lr] = ert[ll];
      ert[rl] = ert[rr];
      rt[lr] = rt[ll];
      rt[rl] = rt[rr];
      rsat[lr] = rsat[ll];
      rsat[rl] = rsat[rr];
      vx[lr] = vx[ll];
      vx[rl] = vx[rr];
      vy[lr] = vy[ll];
      vy[rl] = vy[rr];
      pi[lr] = pi[ll];
      pi[rl] = pi[rr];
      mont[lr] = mont[ll];
      mont[rl] = mont[rr];
      th[lr] = th[ll];
      th[rl] = th[rr];
      z[lr] = z[ll];
      z[rl] = z[rr];
      kh[lr] = kh[ll];
      kh[rl] = kh[rr];
      kv[lr] = kv[ll];
      kv[rl] = kv[rr];
    }

    /* 2-D fields -- step 1 */
    ll = I2(0,iy);
    lr = I2(nx,iy);
    rr = I2(nx - 1,iy);
    rl = I2(-1,iy);
    thb[lr] = thb[ll];
    thb[rl] = thb[rr];
    vxb[lr] = vxb[ll];
    vxb[rl] = vxb[rr];
    vyb[lr] = vyb[ll];
    vyb[rl] = vyb[rr];
    topo[lr] = topo[ll];
    topo[rl] = topo[rr];
    rain[lr] = rain[ll];
    rain[rl] = rain[rr];
    srain[lr] = srain[ll];
    srain[rl] = srain[rr];
    prain[lr] = prain[ll];
    prain[rl] = prain[rr];
    erain[lr] = erain[ll];
    erain[rl] = erain[rr];
    rain1[lr] = rain1[ll];
    rain1[rl] = rain1[rr];
    rain2s[lr] = rain2s[ll];
    rain2s[rl] = rain2s[rr];
    rain2t[lr] = rain2t[ll];
    rain2t[rl] = rain2t[rr];
    eflux[lr] = eflux[ll];
    eflux[rl] = eflux[rr];
    rflux[lr] = rflux[ll];
    rflux[rl] = rflux[rr];
    vxflux[lr] = vxflux[ll];
    vxflux[rl] = vxflux[rr];
    vyflux[lr] = vyflux[ll];
    vyflux[rl] = vyflux[rr];
    fluxthrot[lr] = fluxthrot[ll];
    fluxthrot[rl] = fluxthrot[rr];
    cvcaug[lr] = cvcaug[ll];
    cvcaug[rl] = cvcaug[rr];
    ctopdeep[lr] = ctopdeep[ll];
    ctopdeep[rl] = ctopdeep[rr];
    ctopshallow[lr] = ctopshallow[ll];
    ctopshallow[rl] = ctopshallow[rr];
    convthrot[lr] = convthrot[ll];
    convthrot[rl] = convthrot[rr];
    srftmp[lr] = srftmp[ll];
    srftmp[rl] = srftmp[rr];
    swater[lr] = swater[ll];
    swater[rl] = swater[rr];
    landfrac[lr] = landfrac[ll];
    landfrac[rl] = landfrac[rr];
    sthb[lr] = sthb[ll];
    sthb[rl] = sthb[rr];

    /* 2-D fields -- step 2 */
    ll = I2(1,iy);
    lr = I2(nx + 1,iy);
    rr = I2(nx - 2,iy);
    rl = I2(-2,iy);
    thb[lr] = thb[ll];
    thb[rl] = thb[rr];
    vxb[lr] = vxb[ll];
    vxb[rl] = vxb[rr];
    vyb[lr] = vyb[ll];
    vyb[rl] = vyb[rr];
    topo[lr] = topo[ll];
    topo[rl] = topo[rr];
    rain[lr] = rain[ll];
    rain[rl] = rain[rr];
    srain[lr] = srain[ll];
    srain[rl] = srain[rr];
    prain[lr] = prain[ll];
    prain[rl] = prain[rr];
    erain[lr] = erain[ll];
    erain[rl] = erain[rr];
    rain1[lr] = rain1[ll];
    rain1[rl] = rain1[rr];
    rain2s[lr] = rain2s[ll];
    rain2s[rl] = rain2s[rr];
    rain2t[lr] = rain2t[ll];
    rain2t[rl] = rain2t[rr];
    eflux[lr] = eflux[ll];
    eflux[rl] = eflux[rr];
    rflux[lr] = rflux[ll];
    rflux[rl] = rflux[rr];
    vxflux[lr] = vxflux[ll];
    vxflux[rl] = vxflux[rr];
    vyflux[lr] = vyflux[ll];
    vyflux[rl] = vyflux[rr];
    fluxthrot[lr] = fluxthrot[ll];
    fluxthrot[rl] = fluxthrot[rr];
    cvcaug[lr] = cvcaug[ll];
    cvcaug[rl] = cvcaug[rr];
    ctopdeep[lr] = ctopdeep[ll];
    ctopdeep[rl] = ctopdeep[rr];
    ctopshallow[lr] = ctopshallow[ll];
    ctopshallow[rl] = ctopshallow[rr];
    convthrot[lr] = convthrot[ll];
    convthrot[rl] = convthrot[rr];
    srftmp[lr] = srftmp[ll];
    srftmp[rl] = srftmp[rr];
    swater[lr] = swater[ll];
    swater[rl] = swater[rr];
    landfrac[lr] = landfrac[ll];
    landfrac[rl] = landfrac[rr];
    sthb[lr] = sthb[ll];
    sthb[rl] = sthb[rr];
  }

  if (channel) {

    /* extend by mirror image across walls in channel case */
    for (ix = -2; ix <= nx + 1; ix++) {
      for (ith = 0; ith < nth; ith++) {

	/* step 1 */
	ll = I(ix,ny - 1,ith);
	lr = I(ix,ny,ith);
	rr = I(ix,0,ith);
	rl = I(ix,-1,ith);
	eta[lr] = eta[ll];
	eta[rl] = eta[rr];
	ux[lr] = ux[ll];
	ux[rl] = ux[rr];
	uy[lr] = -uy[ll];
	uy[rl] = -uy[rr];
	fx[lr] = fx[ll];
	fx[rl] = fx[rr];
	fy[lr] = fy[ll];
	fy[rl] = fy[rr];
	sth[lr] = sth[ll];
	sth[rl] = sth[rr];
	sthe[lr] = sthe[ll];
	sthe[rl] = sthe[rr];
	sthe_cu[lr] = sthe_cu[ll];
	sthe_cu[rl] = sthe_cu[rr];
	sthe_rad[lr] = sthe_rad[ll];
	sthe_rad[rl] = sthe_rad[rr];
	srt[lr] = srt[ll];
	srt[rl] = srt[rr];
	ssig[lr] = ssig[ll];
	ssig[rl] = ssig[rr];
	the[lr] = the[ll];
	the[rl] = the[rr];
	thes[lr] = thes[ll];
	thes[rl] = thes[rr];
	ert[lr] = ert[ll];
	ert[rl] = ert[rr];
	rt[lr] = rt[ll];
	rt[rl] = rt[rr];
	rsat[lr] = rsat[ll];
	rsat[rl] = rsat[rr];
	vx[lr] = vx[ll];
	vx[rl] = vx[rr];
	vy[lr] = -vy[ll];
	vy[rl] = -vy[rr];
	pi[lr] = pi[ll];
	pi[rl] = pi[rr];
	mont[lr] = mont[ll];
	mont[rl] = mont[rr];
	th[lr] = th[ll];
	th[rl] = th[rr];
	z[lr] = z[ll];
	z[rl] = z[rr];
	kh[lr] = kh[ll];
	kh[rl] = kh[rr];
	kv[lr] = kv[ll];
	kv[rl] = kv[rr];

	/* step 2 */
	ll = I(ix,ny - 2,ith);
	lr = I(ix,ny + 1,ith);
	rr = I(ix,1,ith);
	rl = I(ix,-2,ith);
	eta[lr] = eta[ll];
	eta[rl] = eta[rr];
	ux[lr] = ux[ll];
	ux[rl] = ux[rr];
	uy[lr] = -uy[ll];
	uy[rl] = -uy[rr];
	fx[lr] = fx[ll];
	fx[rl] = fx[rr];
	fy[lr] = fy[ll];
	fy[rl] = fy[rr];
	sth[lr] = sth[ll];
	sth[rl] = sth[rr];
	sthe[lr] = sthe[ll];
	sthe[rl] = sthe[rr];
	sthe_cu[lr] = sthe_cu[ll];
	sthe_cu[rl] = sthe_cu[rr];
	sthe_rad[lr] = sthe_rad[ll];
	sthe_rad[rl] = sthe_rad[rr];
	srt[lr] = srt[ll];
	srt[rl] = srt[rr];
	ssig[lr] = ssig[ll];
	ssig[rl] = ssig[rr];
	the[lr] = the[ll];
	the[rl] = the[rr];
	thes[lr] = thes[ll];
	thes[rl] = thes[rr];
	ert[lr] = ert[ll];
	ert[rl] = ert[rr];
	rt[lr] = rt[ll];
	rt[rl] = rt[rr];
	rsat[lr] = rsat[ll];
	rsat[rl] = rsat[rr];
	vx[lr] = vx[ll];
	vx[rl] = vx[rr];
	vy[lr] = -vy[ll];
	vy[rl] = -vy[rr];
	pi[lr] = pi[ll];
	pi[rl] = pi[rr];
	mont[lr] = mont[ll];
	mont[rl] = mont[rr];
	th[lr] = th[ll];
	th[rl] = th[rr];
	z[lr] = z[ll];
	z[rl] = z[rr];
	kh[lr] = kh[ll];
	kh[rl] = kh[rr];
	kv[lr] = kv[ll];
	kv[rl] = kv[rr];
      }

      /* 2-D fields -- step 1 */
      ll = I2(ix,ny - 1);
      lr = I2(ix,ny);
      rr = I2(ix,0);
      rl = I2(ix,-1);
      thb[lr] = thb[ll];
      thb[rl] = thb[rr];
      vxb[lr] = vxb[ll];
      vxb[rl] = vxb[rr];
      vyb[lr] = -vyb[ll];
      vyb[rl] = -vyb[rr];
      topo[lr] = topo[ll];
      topo[rl] = topo[rr];
      rain[lr] = rain[ll];
      rain[rl] = rain[rr];
      srain[lr] = srain[ll];
      srain[rl] = srain[rr];
      prain[lr] = prain[ll];
      prain[rl] = prain[rr];
      erain[lr] = erain[ll];
      erain[rl] = erain[rr];
      rain1[lr] = rain1[ll];
      rain1[rl] = rain1[rr];
      rain2s[lr] = rain2s[ll];
      rain2s[rl] = rain2s[rr];
      rain2t[lr] = rain2t[ll];
      rain2t[rl] = rain2t[rr];
      eflux[lr] = eflux[ll];
      eflux[rl] = eflux[rr];
      rflux[lr] = rflux[ll];
      rflux[rl] = rflux[rr];
      vxflux[lr] = vxflux[ll];
      vxflux[rl] = vxflux[rr];
      vyflux[lr] = -vyflux[ll];
      vyflux[rl] = -vyflux[rr];
      fluxthrot[lr] = fluxthrot[ll];
      fluxthrot[rl] = fluxthrot[rr];
      cvcaug[lr] = cvcaug[ll];
      cvcaug[rl] = cvcaug[rr];
      ctopdeep[lr] = ctopdeep[ll];
      ctopdeep[rl] = ctopdeep[rr];
      ctopshallow[lr] = ctopshallow[ll];
      ctopshallow[rl] = ctopshallow[rr];
      convthrot[lr] = convthrot[ll];
      convthrot[rl] = convthrot[rr];
      srftmp[lr] = srftmp[ll];
      srftmp[rl] = srftmp[rr];
      swater[lr] = swater[ll];
      swater[rl] = swater[rr];
      landfrac[lr] = landfrac[ll];
      landfrac[rl] = landfrac[rr];
      sthb[lr] = sthb[ll];
      sthb[rl] = sthb[rr];

      /* 2-D fields -- step 2 */
      ll = I2(ix,ny - 2);
      lr = I2(ix,ny + 1);
      rr = I2(ix,1);
      rl = I2(ix,-2);
      thb[lr] = thb[ll];
      thb[rl] = thb[rr];
      vxb[lr] = vxb[ll];
      vxb[rl] = vxb[rr];
      vyb[lr] = -vyb[ll];
      vyb[rl] = -vyb[rr];
      topo[lr] = topo[ll];
      topo[rl] = topo[rr];
      rain[lr] = rain[ll];
      rain[rl] = rain[rr];
      srain[lr] = srain[ll];
      srain[rl] = srain[rr];
      prain[lr] = prain[ll];
      prain[rl] = prain[rr];
      erain[lr] = erain[ll];
      erain[rl] = erain[rr];
      rain1[lr] = rain1[ll];
      rain1[rl] = rain1[rr];
      rain2s[lr] = rain2s[ll];
      rain2s[rl] = rain2s[rr];
      rain2t[lr] = rain2t[ll];
      rain2t[rl] = rain2t[rr];
      eflux[lr] = eflux[ll];
      eflux[rl] = eflux[rr];
      rflux[lr] = rflux[ll];
      rflux[rl] = rflux[rr];
      vxflux[lr] = vxflux[ll];
      vxflux[rl] = vxflux[rr];
      vyflux[lr] = -vyflux[ll];
      vyflux[rl] = -vyflux[rr];
      fluxthrot[lr] = fluxthrot[ll];
      fluxthrot[rl] = fluxthrot[rr];
      cvcaug[lr] = cvcaug[ll];
      cvcaug[rl] = cvcaug[rr];
      ctopdeep[lr] = ctopdeep[ll];
      ctopdeep[rl] = ctopdeep[rr];
      ctopshallow[lr] = ctopshallow[ll];
      ctopshallow[rl] = ctopshallow[rr];
      convthrot[lr] = convthrot[ll];
      convthrot[rl] = convthrot[rr];
      srftmp[lr] = srftmp[ll];
      srftmp[rl] = srftmp[rr];
      swater[lr] = swater[ll];
      swater[rl] = swater[rr];
      landfrac[lr] = landfrac[ll];
      landfrac[rl] = landfrac[rr];
      sthb[lr] = sthb[ll];
      sthb[rl] = sthb[rr];
    }
  }
  else {

    /* extend periodically in y direction in non-channel case */
    for (ix = -2; ix <= nx + 1; ix++) {
      for (ith = 0; ith < nth; ith++) {

	/* step 1 */
	ll = I(ix,0,ith);
	lr = I(ix,ny,ith);
	rr = I(ix,ny - 1,ith);
	rl = I(ix,-1,ith);
	eta[lr] = eta[ll];
	eta[rl] = eta[rr];
	ux[lr] = ux[ll];
	ux[rl] = ux[rr];
	uy[lr] = uy[ll];
	uy[rl] = uy[rr];
	fx[lr] = fx[ll];
	fx[rl] = fx[rr];
	fy[lr] = fy[ll];
	fy[rl] = fy[rr];
	sth[lr] = sth[ll];
	sth[rl] = sth[rr];
	sthe[lr] = sthe[ll];
	sthe[rl] = sthe[rr];
	sthe_cu[lr] = sthe_cu[ll];
	sthe_cu[rl] = sthe_cu[rr];
	sthe_rad[lr] = sthe_rad[ll];
	sthe_rad[rl] = sthe_rad[rr];
	srt[lr] = srt[ll];
	srt[rl] = srt[rr];
	ssig[lr] = ssig[ll];
	ssig[rl] = ssig[rr];
	the[lr] = the[ll];
	the[rl] = the[rr];
	thes[lr] = thes[ll];
	thes[rl] = thes[rr];
	ert[lr] = ert[ll];
	ert[rl] = ert[rr];
	rt[lr] = rt[ll];
	rt[rl] = rt[rr];
	rsat[lr] = rsat[ll];
	rsat[rl] = rsat[rr];
	vx[lr] = vx[ll];
	vx[rl] = vx[rr];
	vy[lr] = vy[ll];
	vy[rl] = vy[rr];
	pi[lr] = pi[ll];
	pi[rl] = pi[rr];
	mont[lr] = mont[ll];
	mont[rl] = mont[rr];
	th[lr] = th[ll];
	th[rl] = th[rr];
	z[lr] = z[ll];
	z[rl] = z[rr];
	kh[lr] = kh[ll];
	kh[rl] = kh[rr];
	kv[lr] = kv[ll];
	kv[rl] = kv[rr];

	/* step 2 */
	ll = I(ix,1,ith);
	lr = I(ix,ny + 1,ith);
	rr = I(ix,ny - 2,ith);
	rl = I(ix,-2,ith);
	eta[lr] = eta[ll];
	eta[rl] = eta[rr];
	ux[lr] = ux[ll];
	ux[rl] = ux[rr];
	uy[lr] = uy[ll];
	uy[rl] = uy[rr];
	fx[lr] = fx[ll];
	fx[rl] = fx[rr];
	fy[lr] = fy[ll];
	fy[rl] = fy[rr];
	sth[lr] = sth[ll];
	sth[rl] = sth[rr];
	sthe[lr] = sthe[ll];
	sthe[rl] = sthe[rr];
	sthe_cu[lr] = sthe_cu[ll];
	sthe_cu[rl] = sthe_cu[rr];
	sthe_rad[lr] = sthe_rad[ll];
	sthe_rad[rl] = sthe_rad[rr];
	srt[lr] = srt[ll];
	srt[rl] = srt[rr];
	ssig[lr] = ssig[ll];
	ssig[rl] = ssig[rr];
	the[lr] = the[ll];
	the[rl] = the[rr];
	thes[lr] = thes[ll];
	thes[rl] = thes[rr];
	ert[lr] = ert[ll];
	ert[rl] = ert[rr];
	rt[lr] = rt[ll];
	rt[rl] = rt[rr];
	rsat[lr] = rsat[ll];
	rsat[rl] = rsat[rr];
	vx[lr] = vx[ll];
	vx[rl] = vx[rr];
	vy[lr] = vy[ll];
	vy[rl] = vy[rr];
	pi[lr] = pi[ll];
	pi[rl] = pi[rr];
	mont[lr] = mont[ll];
	mont[rl] = mont[rr];
	th[lr] = th[ll];
	th[rl] = th[rr];
	z[lr] = z[ll];
	z[rl] = z[rr];
	kh[lr] = kh[ll];
	kh[rl] = kh[rr];
	kv[lr] = kv[ll];
	kv[rl] = kv[rr];
      }

      /* 2-D fields -- step 1 */
      ll = I2(ix,0);
      lr = I2(ix,ny);
      rr = I2(ix,ny - 1);
      rl = I2(ix,-1);
      thb[lr] = thb[ll];
      thb[rl] = thb[rr];
      vxb[lr] = vxb[ll];
      vxb[rl] = vxb[rr];
      vyb[lr] = vyb[ll];
      vyb[rl] = vyb[rr];
      topo[lr] = topo[ll];
      topo[rl] = topo[rr];
      rain[lr] = rain[ll];
      rain[rl] = rain[rr];
      srain[lr] = srain[ll];
      srain[rl] = srain[rr];
      prain[lr] = prain[ll];
      prain[rl] = prain[rr];
      erain[lr] = erain[ll];
      erain[rl] = erain[rr];
      rain1[lr] = rain1[ll];
      rain1[rl] = rain1[rr];
      rain2s[lr] = rain2s[ll];
      rain2s[rl] = rain2s[rr];
      rain2t[lr] = rain2t[ll];
      rain2t[rl] = rain2t[rr];
      eflux[lr] = eflux[ll];
      eflux[rl] = eflux[rr];
      rflux[lr] = rflux[ll];
      rflux[rl] = rflux[rr];
      vxflux[lr] = vxflux[ll];
      vxflux[rl] = vxflux[rr];
      vyflux[lr] = vyflux[ll];
      vyflux[rl] = vyflux[rr];
      fluxthrot[lr] = fluxthrot[ll];
      fluxthrot[rl] = fluxthrot[rr];
      cvcaug[lr] = cvcaug[ll];
      cvcaug[rl] = cvcaug[rr];
      ctopdeep[lr] = ctopdeep[ll];
      ctopdeep[rl] = ctopdeep[rr];
      ctopshallow[lr] = ctopshallow[ll];
      ctopshallow[rl] = ctopshallow[rr];
      convthrot[lr] = convthrot[ll];
      convthrot[rl] = convthrot[rr];
      srftmp[lr] = srftmp[ll];
      srftmp[rl] = srftmp[rr];
      swater[lr] = swater[ll];
      swater[rl] = swater[rr];
      landfrac[lr] = landfrac[ll];
      landfrac[rl] = landfrac[rr];
      sthb[lr] = sthb[ll];
      sthb[rl] = sthb[rr];

      /* 2-D fields -- step 2 */
      ll = I2(ix,1);
      lr = I2(ix,ny + 1);
      rr = I2(ix,ny - 2);
      rl = I2(ix,-2);
      thb[lr] = thb[ll];
      thb[rl] = thb[rr];
      vxb[lr] = vxb[ll];
      vxb[rl] = vxb[rr];
      vyb[lr] = vyb[ll];
      vyb[rl] = vyb[rr];
      topo[lr] = topo[ll];
      topo[rl] = topo[rr];
      rain[lr] = rain[ll];
      rain[rl] = rain[rr];
      srain[lr] = srain[ll];
      srain[rl] = srain[rr];
      prain[lr] = prain[ll];
      prain[rl] = prain[rr];
      erain[lr] = erain[ll];
      erain[rl] = erain[rr];
      rain1[lr] = rain1[ll];
      rain1[rl] = rain1[rr];
      rain2s[lr] = rain2s[ll];
      rain2s[rl] = rain2s[rr];
      rain2t[lr] = rain2t[ll];
      rain2t[rl] = rain2t[rr];
      eflux[lr] = eflux[ll];
      eflux[rl] = eflux[rr];
      rflux[lr] = rflux[ll];
      rflux[rl] = rflux[rr];
      vxflux[lr] = vxflux[ll];
      vxflux[rl] = vxflux[rr];
      vyflux[lr] = vyflux[ll];
      vyflux[rl] = vyflux[rr];
      fluxthrot[lr] = fluxthrot[ll];
      fluxthrot[rl] = fluxthrot[rr];
      cvcaug[lr] = cvcaug[ll];
      cvcaug[rl] = cvcaug[rr];
      ctopdeep[lr] = ctopdeep[ll];
      ctopdeep[rl] = ctopdeep[rr];
      ctopshallow[lr] = ctopshallow[ll];
      ctopshallow[rl] = ctopshallow[rr];
      convthrot[lr] = convthrot[ll];
      convthrot[rl] = convthrot[rr];
      srftmp[lr] = srftmp[ll];
      srftmp[rl] = srftmp[rr];
      swater[lr] = swater[ll];
      swater[rl] = swater[rr];
      landfrac[lr] = landfrac[ll];
      landfrac[rl] = landfrac[rr];
      sthb[lr] = sthb[ll];
      sthb[rl] = sthb[rr];
    }
  }

  /* extend in sigma direction */
  for (ix = -2; ix <= nx + 1; ix++) {
    for (iy = -2; iy <= ny + 1; iy++) {
      ib = I(ix,iy,-1);
      ib1 = I(ix,iy,0);
      ib2 = I(ix,iy,1);
      it = I(ix,iy,nth);
      it1 = I(ix,iy,nth - 1);
      it2 = I(ix,iy,nth - 2);

      /* linear extrapolation of dependent variables */
      eta[ib] = eta[ib1];
      eta[it] = eta[it1];
      ux[ib] = ux[ib1];
      ux[it] = ux[it1];
      uy[ib] = uy[ib1];
      uy[it] = uy[it1];
      ert[ib] = ert[ib1];
      ert[it] = ert[it1];
      th[ib] = 2.*th[ib1] - th[ib2];
      th[it] = 2.*th[it1] - th[it2];
      ssig[ib] = -ssig[ib1];
      ssig[it] = -ssig[it1];
    }
  }
}





/* initialize */
void step_init(double time)
{
  int n3, n2, nthy, ne3, ne2;
  int i, i2, ix, iy, ith, k;

  /* get parameters needed -- assume main program reads file */
  nx = params_int("nx");
  ny = params_int("ny");
  nth = params_int("nth");
  xbase = params_float("xbase");
  ybase = params_float("ybase");
  thb0 = params_float("thb0");
  tht0 = params_float("tht0");
  dx = params_float("dx");
  dy = params_float("dy");
  dt = params_float("dt");
  ptop = params_float("ptop");
  f0 = params_float("f0");
  beta = params_float("beta");
  kappa = params_float("kappa");
  spngmag = params_float("spngmag");
  spngy = params_float("spngy");
  spngsig = params_float("spngsig");
  spngy = params_float("spngy");
  channel = params_int("channel");
  ckh = params_float("ckh");
  ckv = params_float("ckv");
  lambda = params_float("lambda");
  urelax = params_float("urelax");

  /* compute auxiliary variables */
  dsig = 1./nth;
  dh = 0.5*dt;

  /* allocate buffer space */
  n3 = (nx + 4)*(ny + 4)*(nth + 2);
  n2 = (nx + 4)*(ny + 4);
  nthy = ny*nth;

  /* 3-D primary fields */
  vx = grid_alloc(n3);
  vy = grid_alloc(n3);
  eta = grid_alloc(n3);
  the = grid_alloc(n3);
  thes = grid_alloc(n3);
  rt = grid_alloc(n3);
  rsat = grid_alloc(n3);
  pi = grid_alloc(n3);
  mont = grid_alloc(n3);
  z = grid_alloc(n3);
  th = grid_alloc(n3);
  ux = grid_alloc(n3);
  uy = grid_alloc(n3);
  ert = grid_alloc(n3);
  kh = grid_alloc(n3);
  kv = grid_alloc(n3);

  /* 3-D flux fields */
  fssigeta = grid_alloc(n3);
  fvxux = grid_alloc(n3);
  fvyux = grid_alloc(n3);
  fssigux = grid_alloc(n3);
  fvxuy = grid_alloc(n3);
  fvyuy = grid_alloc(n3);
  fssiguy = grid_alloc(n3);
  fvxert = grid_alloc(n3);
  fvyert = grid_alloc(n3);
  fssigert = grid_alloc(n3);

  /* 3-D source fields */
  fx = grid_alloc(n3);
  fy = grid_alloc(n3);
  sth = grid_alloc(n3);
  sthe = grid_alloc(n3);
  sthe_cu = grid_alloc(n3);
  sthe_rad = grid_alloc(n3);
  srt = grid_alloc(n3);
  ssig = grid_alloc(n3);

  /* 2-D primary fields */
  vxb = grid_alloc(n2);
  vyb = grid_alloc(n2);
  dvx = grid_alloc(n2);
  dvy = grid_alloc(n2);
  thb = grid_alloc(n2);
  thbtemp = grid_alloc(n2);
  srftmp = grid_alloc(n2);
  swater = grid_alloc(n2);
  landfrac = grid_alloc(n2);

  /* 2-D temporary fields */
  sth1 = grid_alloc(n2);
  srt1 = grid_alloc(n2);
  fx1 = grid_alloc(n2);
  fy1 = grid_alloc(n2);

  /* 2-D flux output fields */
  rain = grid_alloc(n2);
  srain = grid_alloc(n2);
  prain = grid_alloc(n2);
  erain = grid_alloc(n2);
  rain1 = grid_alloc(n2);
  rain2s = grid_alloc(n2);
  rain2t = grid_alloc(n2);
  eflux = grid_alloc(n2);
  rflux = grid_alloc(n2);
  vxflux = grid_alloc(n2);
  vyflux = grid_alloc(n2);
  convthrot = grid_alloc(n2);
  fluxthrot = grid_alloc(n2);
  cvcaug = grid_alloc(n2);
  ctopdeep = grid_alloc(n2);
  ctopshallow = grid_alloc(n2);

  /* 2-D input fields */
  topo = grid_alloc(n2);

  /* 2-D source fields */
  sthb = grid_alloc(n2);

  /* Coriolis parameter */
  f = grid_alloc(ny + 4);

  /* temporary vertical profiles */
  sth2 = grid_alloc(nth);
  srt2 = grid_alloc(nth);
  fx2 = grid_alloc(nth);
  fy2 = grid_alloc(nth);

  /* temporary zonally symmetric fields */
  vxm = grid_alloc(nthy);
  vxref = grid_alloc(nthy);

  /* corner buffers */
  ne3 = (nx + 1)*(ny + 1)*(nth + 1);

  /* 3-D primary fields */
  xeta = grid_alloc(ne3);
  xux = grid_alloc(ne3);
  xuy = grid_alloc(ne3);
  xert = grid_alloc(ne3);
  xth = grid_alloc(ne3);
  xpi = grid_alloc(ne3);
  xmont = grid_alloc(ne3);
  xz = grid_alloc(ne3);
  xkh = grid_alloc(ne3);
  xkv = grid_alloc(ne3);

  /* flux fields */
  xfssigeta = grid_alloc(ne3);
  xfvxux = grid_alloc(ne3);
  xfvyux = grid_alloc(ne3);
  xfssigux = grid_alloc(ne3);
  xfvxuy = grid_alloc(ne3);
  xfvyuy = grid_alloc(ne3);
  xfssiguy = grid_alloc(ne3);
  xfvxert = grid_alloc(ne3);
  xfvyert = grid_alloc(ne3);
  xfssigert = grid_alloc(ne3);
  xfvxeta = grid_alloc(ne3);
  xfvyeta = grid_alloc(ne3);

  /* 1-D fields */
  xf = grid_alloc(ny + 1);

  /* auxiliary fields */
  etamin = grid_alloc(nth);
  etamax = grid_alloc(nth);

  /* calculate Coriolis parameter */
  for (iy = -2; iy <= ny + 1; iy++) {
    f[IY(iy)] = f0 + beta*(ybase + dy*iy);
  }
  for (iy = 0; iy <= ny; iy++) {
    xf[JY(iy)] = 0.5*(f[IY(iy - 1)] + f[IY(iy)]);
  }

  /* initialize the source term calculator */
  source_init();

  /* initialize the vertical integrator */
  vert_init();

  /* initialize topography, prognostic, and auxiliary fields */
  init_fields(thb, srftmp, landfrac, topo, swater, eta, vx, vy, ssig, pi,
	      mont, z, th, the, thes, rt, rsat, ux, uy, ert, etamin, etamax,
	      vxb, vyb, rain, srain, prain, erain, rain1, rain2s, rain2t,
	      eflux, rflux, vxflux, vyflux, convthrot, fluxthrot, cvcaug,
	      ctopdeep,ctopshallow);

  /* initialize reference zonal wind */
  for (ith = 0; ith < nth; ith++) {
    for (iy = 0; iy < ny; iy++) {
      k = K(iy,ith);
      vxref[k] = 0.;
      for (ix = 0; ix < nx; ix++) {
	i = I(ix,iy,ith);
	vxref[k] += ux[i]/eta[i];
      }
      vxref[k] = vxref[k]/nx;
    }
  }

  /* get the source terms */
  source_comp(time,thb,srftmp,landfrac,topo,swater,eta,pi,z,mont,th,the,rt,
	      vx,vy,fx,fy,sth,sthe,sthe_cu,sthe_rad,srt,rain1,rain2s,rain2t,
	      rain,srain,erain,eflux,rflux,vxflux,vyflux,
	      convthrot,fluxthrot,cvcaug,ctopdeep,ctopshallow);

  /* extend all fields to encompassing grid */
  c_extend();
}





/* do a time step */
void step_do(double time)
{
  int ix, iy, ith, i, i2, j, j2, im, ip, k, jth;
  int ipp, ipm, imp, imm;
  double sigma, xssig, dxx, dyy, dxy, dss;
  double strainrate, heddy, veddy, xthb, lopart, hipart;
  double etadist;

  /* compute horizontal mean profile of vx */
  for (ith = 0; ith < nth; ith++) {
    for (iy = 0; iy < ny; iy++) {
      k = K(iy,ith);
      vxm[k] = 0.;
      for (ix = 0; ix < nx; ix++) {
	vxm[k] += vx[I(ix,iy,ith)];
      }
      vxm[k] = vxm[k]/nx;
    }
  }

  /* compute fluxes on center grid */
  for (ix = -1; ix <= nx; ix++) {
    for (ith = -1; ith <= nth; ith++) {
      for (iy = -1; iy <= ny; iy++) {
	i = I(ix,iy,ith);
	fssigeta[i] = ssig[i]*eta[i];
	fvxux[i] = ux[i]*ux[i]/eta[i];
	fvyux[i] = uy[i]*ux[i]/eta[i];
	fssigux[i] = ssig[i]*ux[i];
	fvxuy[i] = fvyux[i];
	fvyuy[i] = uy[i]*uy[i]/eta[i];
	fssiguy[i] = ssig[i]*uy[i];
	fvxert[i] = ux[i]*ert[i]/eta[i];
	fvyert[i] = uy[i]*ert[i]/eta[i];
	fssigert[i] = ssig[i]*ert[i];
      }
    }
  }

  /* do half step part of Lax-Wendroff -- loop over J grid */
  for (ix = 0; ix <= nx; ix++) {
    for (iy = 0; iy <= ny; iy++) {
      for (ith = nth; ith >= 0; ith--) {
	j = J(ix,iy,ith);

	/* compute xth */
	xthb = AV2(ix,iy,thb);
	xth[j] = ith*dsig*(tht0 - xthb) + xthb;

	/* 3-D fields */
	xux[j] = AV(ix,iy,ith,ux) +
	  (-DIVX(ix,iy,ith,fvxux) - DIVY(ix,iy,ith,fvyux) -
	   DIVTH(ix,iy,ith,fssigux) +
	   AV(ix,iy,ith,eta)*(- DIVX(ix,iy,ith,mont)
			      + AV(ix,iy,ith,pi)*DIVX(ix,iy,ith,th)
			      + AV(ix,iy,ith,fx)) +
	   xf[JY(iy)]*AV(ix,iy,ith,uy))*dh;
	
	if (channel && (iy == 0 || iy == ny)) {
	  xuy[j] = 0.;
	}
	else {
	  xuy[j] = AV(ix,iy,ith,uy) +
	    (-DIVX(ix,iy,ith,fvxuy) - DIVY(ix,iy,ith,fvyuy) -
	     DIVTH(ix,iy,ith,fssiguy) +
	     AV(ix,iy,ith,eta)*(- DIVY(ix,iy,ith,mont)
				+ AV(ix,iy,ith,pi)*DIVY(ix,iy,ith,th)
				+ AV(ix,iy,ith,fy)) -
	     xf[JY(iy)]*AV(ix,iy,ith,ux))*dh;
	}
	
	xeta[j] = AV(ix,iy,ith,eta) +
	  (-DIVX(ix,iy,ith,ux) - DIVY(ix,iy,ith,uy) -
	   DIVTH(ix,iy,ith,fssigeta))*dh;

	/* don't let eta go out of range */
	jth = ith < nth ? ith : nth - 1;
	if (xeta[j] < etamin[jth]) xeta[j] = etamin[jth];
	if (xeta[j] > etamax[jth]) xeta[j] = etamax[jth];

	xert[j] = AV(ix,iy,ith,ert) +
	  (-DIVX(ix,iy,ith,fvxert) - DIVY(ix,iy,ith,fvyert) -
	   DIVTH(ix,iy,ith,fssigert) +
	   AV(ix,iy,ith,eta)*AV(ix,iy,ith,srt))*dh;

	/* don't allow negative xert */
	if (xert[j] < 0.) xert[j] = 0.;
      }

      /* vertical integration -- corner input and output */
      j2 = J2(ix,iy);
      j = J(ix,iy,0);
      vert_jcomp(AV2(ix,iy,thb),AV2(ix,iy,topo),&xeta[j],&xpi[j],
		 &xmont[j],&xz[j]);
    }
  }

  /* compute fluxes on corner grid */
  for (ix = 0; ix <= nx; ix++) {
    for (iy = 0; iy <= ny; iy++) {
      for (ith = 0; ith <= nth; ith++) {
	j = J(ix,iy,ith);

	/* zero vertical vel on top, bottom -- otherwise, corner value */
	if (ith == 0 || ith == nth) {
	  xssig = 0.;
	}
	else {
	  xssig = AV(ix,iy,ith,ssig);
	}

	/* compute sponge part of horizontal eddy mixing coefficients */
	sigma = ith*dsig;
	heddy = 0.25*(sigma - spngsig)/(1. - spngsig);
	if (heddy < 0.) heddy = 0.;
	veddy = 0.;

	/* if channel, make sponge layers near channel walls as well */
	if (channel) {
	  hipart = 0.25*(YC(iy) - YC(ny) + spngy)/spngy;
	  if (hipart < 0.) hipart = 0.;
	  lopart = 0.25*(YC(0) + spngy - YC(iy))/spngy;
	  if (lopart < 0.) lopart = 0.;
	  heddy += spngmag*(lopart + hipart);
	}
	if (heddy < 0.) heddy = 0.;

	/* strain rate part for both horiz and vert coefficients */
	dxx = DIVX(ix,iy,ith,vx);
	dyy = DIVY(ix,iy,ith,vy);
	dxy = 0.5*(DIVY(ix,iy,ith,vx) + DIVX(ix,iy,ith,vy));
	dss = DIVTH(ix,iy,ith,ssig);
	strainrate = pow(dxx*dxx + dyy*dyy + 2.*dxy*dxy,0.5);
	heddy += ckh*strainrate*dt;
	veddy += ckv*fabs(dss)*dt;

	/* put in dx*dy part, limiting the value to prevent instability */
	if (heddy > 0.25) heddy = 0.25;
	if (veddy > 0.25) veddy = 0.25;
	heddy *= xeta[j]*(dx*dy/dt);
	veddy *= xeta[j]*(dsig*dsig/dt);
	xkh[j] = heddy;
	xkv[j] = veddy;

	/* define the fluxes */
	xfvxux[j] = xux[j]*xux[j]/xeta[j] - heddy*DIVX(ix,iy,ith,vx);
	xfvyux[j] = xuy[j]*xux[j]/xeta[j] - heddy*DIVY(ix,iy,ith,vx);
	xfssigux[j] = xssig*xux[j] - veddy*DIVTH(ix,iy,ith,vx);
	xfvxuy[j] = xux[j]*xuy[j]/xeta[j] - heddy*DIVX(ix,iy,ith,vy);
	xfvyuy[j] = xuy[j]*xuy[j]/xeta[j] - heddy*DIVY(ix,iy,ith,vy);
	xfssiguy[j] = xssig*xuy[j] - veddy*DIVTH(ix,iy,ith,vy);
	xfvxert[j] = xux[j]*xert[j]/xeta[j] - heddy*DIVX(ix,iy,ith,rt);
	xfvyert[j] = xuy[j]*xert[j]/xeta[j] - heddy*DIVY(ix,iy,ith,rt);
	xfssigert[j] = xssig*xert[j] - veddy*DIVTH(ix,iy,ith,rt);
	xfvxeta[j] = xux[j];
	xfvyeta[j] = xuy[j];
	xfssigeta[j] = xssig*xeta[j];

	/* channel sidewall conditions -- free slip, insulating */
	if (channel && (iy == 0 || iy == ny)) {
	  xfvyux[j] = 0.;
	  xfvxuy[j] = 0.;
	  xfvyuy[j] = 0.;
	  xfssiguy[j] = 0.;
	  xfvyert[j] = 0.;
	  xfvyeta[j] = 0.;
	}

	/* top and bottom conditions -- free slip, insulating */
	if (ith == 0 || ith == nth) {
	  xfssigeta[j] = 0.;
	  xfssigux[j] = 0.;
	  xfssiguy[j] = 0.;
	  xfssigert[j] = 0.;
	}
      }
    }
  }

  /* do full step part of Lax-Wendroff -- loop over inner I grid */
  for (ix = 0; ix < nx; ix++) {
    for (iy = 0; iy < ny; iy++) {
      etadist = 0.;
      for (ith = nth - 1; ith >= 0; ith--) {
	i = I(ix,iy,ith);
	k = K(iy,ith);

	/* 3-D fields */
	eta[i] = eta[i] + etadist +
	  (-XDIVX(ix,iy,ith,xfvxeta) - XDIVY(ix,iy,ith,xfvyeta) -
	   XDIVTH(ix,iy,ith,xfssigeta))*dt;

	/* deal with out of range eta */
	if (eta[i] < etamin[ith]) {
	  etadist = eta[i] - etamin[ith];
	  eta[i] = etamin[ith];
	}
	else if (eta[i] > etamax[ith]) {
	  etadist = eta[i] - etamax[ith];
	  eta[i] = etamax[ith];
	}
	else {
	  etadist = 0.;
	}

	ux[i] = ux[i] +
	  (-XDIVX(ix,iy,ith,xfvxux) - XDIVY(ix,iy,ith,xfvyux) -
	   XDIVTH(ix,iy,ith,xfssigux) +
	   XAV(ix,iy,ith,xeta)*(- XDIVX(ix,iy,ith,xmont)
				+ XAV(ix,iy,ith,xpi)*XDIVX(ix,iy,ith,xth)
				+ fx[i] + urelax*(vxref[k] - vxm[k])) +
	   f[IY(iy)]*XAV(ix,iy,ith,xuy))*dt;
	
	if (iy == 0 || iy == ny - 1) {
	  uy[i] = 0.;
	}
	else {
	  uy[i] = uy[i] +
	    (-XDIVX(ix,iy,ith,xfvxuy) - XDIVY(ix,iy,ith,xfvyuy) -
	     XDIVTH(ix,iy,ith,xfssiguy) +
	     XAV(ix,iy,ith,xeta)*(- XDIVY(ix,iy,ith,xmont)
				  + XAV(ix,iy,ith,xpi)*XDIVY(ix,iy,ith,xth)
				  + fy[i]) -
	     f[IY(iy)]*XAV(ix,iy,ith,xux))*dt;
	}
	
	ert[i] = ert[i] +
	  (-XDIVX(ix,iy,ith,xfvxert) - XDIVY(ix,iy,ith,xfvyert) -
	   XDIVTH(ix,iy,ith,xfssigert) +
	   XAV(ix,iy,ith,xeta)*srt[i])*dt;

	/* don't allow negative ert, even though this might kill water cons */
	if (ert[i] < 0.) ert[i] = 0.;
      }

      /* propagate fault upward if left over eta at bottom of column */
      if (etadist != 0.) {
	for (ith = 0; ith < nth; ith++) {
	  i = I(ix,iy,ith);
	  eta[i] += etadist;
	  if (eta[i] < etamin[ith]) {
	    etadist = eta[i] - etamin[ith];
	    eta[i] = etamin[ith];
	  }
	  else if (eta[i] > etamax[ith]) {
	    etadist = eta[i] - etamax[ith];
	    eta[i] = etamax[ith];
	  }
	  else {
	    etadist = 0.;
	  }
	}

	/* if there is still etadist left over at the top, quit */
	if (etadist != 0) {
	  fprintf(stderr,"step: etadist = %g != 0 at %d %d %g\n",
		etadist,ix,iy,time);
	  for (ith = 0; ith < nth; ith++) {
	    i = I(ix,iy,ith);
	    fprintf(stderr,"  %10.5f <= %10.5f <= %10.5f\n",
		    etamin[ith],eta[i],etamax[ith]);
	  }
	  exit(1);
	}
      }

      /* do vertical hypersmoothing */
      i = I(ix,iy,0);
      vsmooth(nth,&eta[i],lambda);
      vsmooth(nth,&ert[i],lambda);
      vsmooth(nth,&ux[i],lambda);
      vsmooth(nth,&uy[i],lambda);
    }
  }

  /* get intensive variables and store eddy mixing coefficients */
  for (ix = 0; ix < nx; ix++) {
    for (iy = 0; iy < ny; iy++) {
      for (ith = 0; ith < nth; ith++) {
	i = I(ix,iy,ith);
	vx[i] = ux[i]/eta[i];
	vy[i] = uy[i]/eta[i];
	rt[i] = ert[i]/eta[i];
	if (rt[i] < 0.) rt[i] = 0.;
	kh[i] = XAV(ix,iy,ith,xkh);
	kv[i] = XAV(ix,iy,ith,xkv);
      }
    }
  }

  /* compute the source terms */
  source_comp(time,thb,srftmp,landfrac,topo,swater,eta,pi,z,mont,th,the,rt,
	      vx,vy,fx,fy,sth,sthe,sthe_cu,sthe_rad,srt,rain1,rain2s,rain2t,
	      rain,srain,erain,eflux,rflux,vxflux,vyflux,
	      convthrot,fluxthrot,cvcaug,ctopdeep,ctopshallow);

  /* extend I grid again after horizontal smoothing on central part of grid */
  c_extend();

  /* set up surface fields in preparation for integrating surface th */
  for (ix = -1; ix <= nx; ix++) {
    for (iy = -1; iy <= ny; iy++) {
      i2 = I2(ix,iy);
      thbtemp[i2] = thb[i2];
      sthb[i2] = SURFEXTRAP(sth,ix,iy);
      vxb[i2] = SURFEXTRAP(vx,ix,iy);
      vyb[i2] = SURFEXTRAP(vy,ix,iy);
    }
  }

  /* do stuff related to the surface as well as vertical profiles */
  for (ix = 0; ix < nx; ix++) {
    for (iy = 0; iy < ny; iy++) {
      i = I(ix,iy,0);
      i2 = I2(ix,iy);

      /* advect surface potential temperature field */
      thb[i2] = thbtemp[i2] +
	(-ADVX(ix,iy,vxb,thbtemp) - ADVY(ix,iy,vyb,thbtemp) + sthb[i2])*dt;

      /* vertical integration -- center input and output */
      vert_icomp(thb[i2],topo[i2],&eta[i],&pi[i],&mont[i],&z[i]);

      /* compute th, the, thes fields */
      for (ith = 0; ith < nth; ith++) {
	i = I(ix,iy,ith);
	if(rt[i] < 0.) {
	  fprintf(stderr,"negative rt at %d %d %d %f\n",ix,iy,ith,time);
	}
	th[i] = thb[i2] + SIGC(ith)*(tht0 - thb[i2]);
	rsat[i] = rs(th[i],pi[i]);
	the[i] = thetaecomp2(th[i],rt[i],rsat[i]);
	thes[i] = thetaecomp2(th[i],rsat[i],rsat[i]);
      }
    }
  }

  /* compute ssig */
  for (ith = 0; ith < nth; ith++) {

    /* first get the dvs over the extended grid */
    for (ix = -1; ix <= nx; ix++) {
      for (iy = -1; iy <= ny; iy++) {
	i2 = I2(ix,iy);
	i = I(ix,iy,ith);
	dvx[i2] = vx[i] - vxb[i2];
	dvy[i2] = vy[i] - vyb[i2];
      }
    }

    /* finally, compute ssig on unextended grid */
    for (ix = 0; ix < nx; ix++) {
      for (iy = 0; iy < ny; iy++) {
	i2 = I2(ix,iy);
	i = I(ix,iy,ith);
	ssig[i] = (sth[i] - (1. - SIGC(ith))*
		   (sthb[i2] + ADVX(ix,iy,dvx,thb) + ADVY(ix,iy,dvy,thb)))/
		   (tht0 - thb[i2]);
      }
    }
  }

  /* do horizontal hypersmoothing */
  hsmooth(nx,ny,nth,channel,eta,kappa);
  hsmooth(nx,ny,nth,channel,ert,kappa);
  hsmooth(nx,ny,nth,channel,ux,kappa);
  hsmooth(nx,ny,nth,channel,uy,kappa);

  /* extend I grid yet again after surface, ssig, and hsmooth */
  c_extend();

}




/* do output to a candis file */
/* write 3-D x-y-z fields */
static void field_xyz(int handle, char *fname, int dimx, int dimy,
		      int dimth, double *fptr)
{
  int i, ix, iy, ith;
  float *lptr;

  /* allocate space */
  if ((lptr = (float*)malloc(sizeof(float)*nx1*ny1*nth)) ==
      (float*)NULL) {
    fprintf(stderr,"Step: Can't get buffer space\n");
    exit(1);
  }

  /* transfer to linear buffer with the positive xy exterior stuff */
  i = 0;
  for (ix = 0; ix < nx1; ix++) {
    for (iy = 0; iy < ny1; iy++) {
      for (ith = 0; ith < nth; ith++) {
	lptr[i++] = fptr[I(ix,iy,ith)];
      }
    }
  }

  /* write the field */
  candisfld(handle,fname,3,dimx,dimy,dimth,0,lptr);

  /* free space */
  free(lptr);
}

/* write 2-D x-y fields */
static void field_xy(int handle, char *fname, int dimx, int dimy,
		     double *fptr)
{
  int i, ix, iy;
  float *lptr;

  /* allocate space */
  if ((lptr = (float*)malloc(sizeof(float)*nx1*ny1)) ==
      (float*)NULL) {
    fprintf(stderr,"Step: Can't get buffer space\n");
    exit(1);
  }

  /* transfer to linear buffer without the exterior stuff */
  i = 0;
  for (ix = 0; ix < nx1; ix++) {
    for (iy = 0; iy < ny1; iy++) {
      lptr[i++] = fptr[I2(ix,iy)];
    }
  }

  /* write the field */
  candisfld(handle,fname,2,dimx,dimy,0,0,lptr);

  /* free space */
  free(lptr);
}

/* one-dimensional y field */
static void field_y(int handle, char *fname, int dimy, double *fptr)
{
  int i, iy;
  float *lptr;

  /* allocate space */
  if ((lptr = (float*)malloc(sizeof(float)*ny1)) == (float*)NULL) {
    fprintf(stderr,"Step: Can't get buffer space\n");
    exit(1);
  }

  /* transfer to linear buffer without the exterior stuff */
  for (iy = 0; iy < ny1; iy++) {
    lptr[iy] = fptr[IY(iy)];
  }

  /* write the field */
  candisfld(handle,fname,1,dimy,0,0,0,lptr);

  /* free space */
  free(lptr);
}

void step_output(char *outfilename, int seqnum, double time, char* progversion)
{
  char label[81];
  int ix, iy, ith, i;
  int chandle, dimx, dimy, dimth;

  /* set up dimensions of output file */
  nx1 = nx + 1;
  ny1 = channel ? ny : ny + 1;

  /* create the output file with the specified name and sequence number */
  chandle = candiscreate(outfilename,seqnum);

  /* make comment */
  strcpy(label,"Isentrope-");
  strncat(label,progversion,60);
  candiscomment(chandle,label);

  /* parameters */
  params_candis_out(chandle);

  /* dimensions */
  dimx = candisdim(chandle,"x",nx1,xbase,dx);
  dimy = candisdim(chandle,"y",ny1,ybase,dy);
  dimth = candisdim(chandle,"sigma",nth,dsig/2.,dsig);

  /* fields */
  dcandisfld(chandle,"time",0,0,0,0,0,&time);
  field_xyz(chandle,"vx",dimx,dimy,dimth,vx);
  field_xyz(chandle,"vy",dimx,dimy,dimth,vy);
  field_xyz(chandle,"eta",dimx,dimy,dimth,eta);
  field_xyz(chandle,"th",dimx,dimy,dimth,th);
  field_xyz(chandle,"the",dimx,dimy,dimth,the);
  field_xyz(chandle,"thes",dimx,dimy,dimth,thes);
  field_xyz(chandle,"rt",dimx,dimy,dimth,rt);
  field_xyz(chandle,"rsat",dimx,dimy,dimth,rsat);
  field_xyz(chandle,"pi",dimx,dimy,dimth,pi);
  field_xyz(chandle,"mont",dimx,dimy,dimth,mont);
  field_xyz(chandle,"z",dimx,dimy,dimth,z);
  field_xyz(chandle,"fx",dimx,dimy,dimth,fx);
  field_xyz(chandle,"fy",dimx,dimy,dimth,fy);
  field_xyz(chandle,"sth",dimx,dimy,dimth,sth);
  field_xyz(chandle,"sthe",dimx,dimy,dimth,sthe);
  field_xyz(chandle,"sthe_cu",dimx,dimy,dimth,sthe_cu);
  field_xyz(chandle,"sthe_rad",dimx,dimy,dimth,sthe_rad);
  field_xyz(chandle,"srt",dimx,dimy,dimth,srt);
  field_xyz(chandle,"ssig",dimx,dimy,dimth,ssig);
  field_xyz(chandle,"kh",dimx,dimy,dimth,kh);
  field_xyz(chandle,"kv",dimx,dimy,dimth,kv);
  field_xy(chandle,"thb",dimx,dimy,thb);
  field_xy(chandle,"topo",dimx,dimy,topo);
  field_xy(chandle,"srftmp",dimx,dimy,srftmp);
  field_xy(chandle,"swater",dimx,dimy,swater);
  field_xy(chandle,"landfrac",dimx,dimy,landfrac);
  field_xy(chandle,"sthb",dimx,dimy,sthb);
  field_xy(chandle,"vxb",dimx,dimy,vxb);
  field_xy(chandle,"vyb",dimx,dimy,vyb);
  field_xy(chandle,"rain",dimx,dimy,rain);
  field_xy(chandle,"srain",dimx,dimy,srain);
  field_xy(chandle,"prain",dimx,dimy,prain);
  field_xy(chandle,"erain",dimx,dimy,erain);
  field_xy(chandle,"rain1",dimx,dimy,rain1);
  field_xy(chandle,"rain2s",dimx,dimy,rain2s);
  field_xy(chandle,"rain2t",dimx,dimy,rain2t);
  field_xy(chandle,"eflux",dimx,dimy,eflux);
  field_xy(chandle,"rflux",dimx,dimy,rflux);
  field_xy(chandle,"vxflux",dimx,dimy,vxflux);
  field_xy(chandle,"vyflux",dimx,dimy,vyflux);
  field_xy(chandle,"convthrot",dimx,dimy,convthrot);
  field_xy(chandle,"fluxthrot",dimx,dimy,fluxthrot);
  field_xy(chandle,"cvcaug",dimx,dimy,cvcaug);
  field_xy(chandle,"ctopdeep",dimx,dimy,ctopdeep);
  field_xy(chandle,"ctopshallow",dimx,dimy,ctopshallow);
  field_y(chandle,"f",dimy,f);

  /* finish the job */
  candiswrite(chandle);
}
