/* thermo.c -- Moist thermodynamics package -- derived from diabat2.
 * Variable naming convection:
 * th: Potential temperature (K)
 * pi: Exner function (J/kg/K)
 * td: Dewpoint temperature (K)
 * the: Equivalent potential temperature (K)
 * rt: Total cloud water (vapor plus droplets) mixing ratio (g/g)
 * rsat: Saturation mixing ratio (g/g)
 * p: Pressure (pa)
 */
#include <stdio.h>
#include <math.h>
#include "thermo.h"

/* thetasat -- assume saturation and calculate theta -- really
 * theta-l */
static TFLOAT thetasat(TFLOAT the,TFLOAT pi,TFLOAT thstart,TINT nloops)
{
  TFLOAT theta,rsval,expfactor,drsdth;
  TINT loop;
  theta = thstart;
  for (loop = 0; loop < nloops; loop++) {
    if (theta < 250.) theta = 250.;
    rsval = rs(theta,pi);
    expfactor = the*exp(-ALFA*rsval);
    drsdth = rs(theta + 1.,pi) - rsval;
    theta -= (theta - expfactor)/(1. + ALFA*expfactor*drsdth);
  }
  if (theta < 250.) theta = 250.;
  return(theta);
}

/* rs - compute saturation mixing ratio (g/g) as function of
 * potential temp (K) and Exner function (J/kg/K)
 * use simple formula
 */
TFLOAT rs(TFLOAT th,TFLOAT pi)
{
  TFLOAT p,es,rsat;
  p = PREF*pow(pi/CP,1./KAPPA);
  es = escalc(th*pi/CP);
  rsat = EPSILON*es/p;
  return(rsat);
}

/* escalc -- compute vapor pressure (pa) from dewpoint (K) -- Teten formula */
TFLOAT escalc(TFLOAT td)
{
  return(611.2*exp(17.67*(td - 273.15)/(td - 29.65)));
}

/* thetacomp2 -- do this computation by a lookup table */
#define ITHEMAX 301
#define IPIMAX 426
#define DTHE 1.
#define DPI 2.
#define THE0 200.
#define PI0 200.

TFLOAT thetacomp2(TFLOAT the,TFLOAT rt,TFLOAT pi)
{
  static TFLOAT v[ITHEMAX][IPIMAX];
  static TINT begin = 1;
  TINT ithe,ipi,jthe,jpi;
  TFLOAT theval,pival,athe,api,othe,opi,theta;
  
  /* fill the lookup table the first time around */
  if (begin) {
    for (ithe = 0; ithe < ITHEMAX; ithe++) {
      for (ipi = 0; ipi < IPIMAX; ipi++) {
	theval = THE0 + ithe*DTHE;
	pival = PI0 + ipi*DPI;
	v[ithe][ipi] = thetasat(theval,pival,theval*exp(-ALFA*rt),5);
      }
    }
    begin = 0;
  }
  
  /* compute test value of theta */
  theta = the*exp(-ALFA*rt);
  
  /* if unsaturated, don't do anything */
  if (rt < rs(theta,pi)) {
  }
  
  /* if saturated, we interpolate */
  else {
    ithe = (the - THE0)/DTHE;
    if (ithe < 0) ithe = 0;
    if (ithe >= ITHEMAX - 1) ithe = ITHEMAX - 2;
    jthe = ithe + 1;
    ipi = (pi - PI0)/DPI;
    if (ipi < 0) ipi = 0;
    if (ipi >= IPIMAX - 1) ipi = IPIMAX - 2;
    jpi = ipi + 1;
    athe = (the - (THE0 + DTHE*ithe))/DTHE;
    othe = 1. - athe;
    api = (pi - (PI0 + DPI*ipi))/DPI;
    opi = 1. - api;
    theta = othe*(opi*v[ithe][ipi] + api*v[ithe][jpi])
      + athe*(opi*v[jthe][ipi] + api*v[jthe][jpi]);
  }
  return(theta);
}

/* thetacomp -- compute potential temperature from equivalent
 * potential temperature, total water mixing ratio, and Exner function.
 */
TFLOAT thetacomp(TFLOAT the,TFLOAT rt,TFLOAT pi)
{
  TFLOAT theta;
  
  /* compute test value of theta */
  theta = the*exp(-ALFA*rt);
  
  /* if unsaturated, don't do anything */
  if (rt < rs(theta,pi)) {
  }
  
  /* if saturated -- we have to do a Newton's method iteration */
  else {
    theta = thetasat(the,pi,theta,5);
  }
  
  /* return with theta value */
  return(theta);
}

/* calculate theta_e */
TFLOAT thetaecomp(TFLOAT th,TFLOAT rt,TFLOAT pi)
{
  return(thetaecomp2(th,rt,rs(th,pi)));
}

/* calculate theta_e given the saturation mixing ratio */
TFLOAT thetaecomp2(TFLOAT th,TFLOAT rt,TFLOAT rsat)
{
  TFLOAT rvap;
  rvap = rt > rsat ? rsat : rt;
  return(th*exp(ALFA*rvap));
}

/* thetaescomp -- compute saturated thetae value given theta (K) and
 * pi (J/kg/K)
 */
TFLOAT thetaescomp(TFLOAT th,TFLOAT pi)
{
  return(th*exp(ALFA*rs(th,pi)));
}

/* gmoist -- compute (d theta/dz) for moist adiabatic parcel */
TFLOAT gmoist(TFLOAT th,TFLOAT pi)
{
  TFLOAT drsdpi,drsdth,rate;
  drsdpi = rs(th,pi + 1.) - rs(th,pi);
  drsdth = rs(th + 1.,pi) - rs(th,pi);
  rate = GEE*ALFA*drsdpi/(1. + th*ALFA*drsdth);
  return(rate);
}

/* pifromp -- compute exner function (J/K/kg) from pressure (Pa) */
TFLOAT pifromp(TFLOAT p)
{
  return(CP*pow(p/PREF,KAPPA));
}

/* pfrompi -- compute pressure (Pa) from exner function (J/K/kg) */
TFLOAT pfrompi(TFLOAT pi)
{
  return(PREF*pow(pi/CP,1./KAPPA));
}

/* rhocomp -- compute density (kg/m^3) from theta (K) and exner (J/K/kg) */
TFLOAT rhocomp(TFLOAT th, TFLOAT pi)
{
  return(PREF*pow(pi/CP,1./KAPPA)/(KAPPA*th*pi));
}
