/* vert.c -- Compute vertical thermodynamics.
 */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "params.h"
#include "vert.h"
#include "thermo.h"
#include "index.h"

static int nth;
static double dsig, ptop, tht0;
static double *epi;

/* initialize parameters */
void vert_init()
{

  /* pick up needed parameters */
  nth = params_int("nth");
  ptop = params_float("ptop");
  tht0 = params_float("tht0");
  dsig = 1./nth;

  /* allocate space for pi on edge grid */
  if ((epi = (double*)malloc(sizeof(double)*(nth + 1))) == (double*)NULL) {
    fprintf(stderr,"Vert: can't get memory\n");
  }
}

/* integrate vertical profiles on centered grid, yielding centered values */
void vert_icomp(double thb, double topo, double *eta, double *pi,
		double *mont, double *z)
{
  int ith;
  double p, pold;

  /* integrate -eta downward to get pi -- values of p on grid edges */
  p = ptop;
  epi[nth] = pifromp(p);
  for (ith = nth - 1; ith >= 0; ith--) {
    pold = p;
    p = pold + GEE*eta[ith]*dsig;
    epi[ith] = pifromp(p);
    pi[ith] = 0.5*(epi[ith] + epi[ith + 1]);
  }

  /* integrate pi upward to get mont */
  mont[0] = thb*epi[0] + GEE*topo;
  for (ith = 1; ith < nth; ith++) {
    mont[ith] = mont[ith - 1] + (tht0 - thb)*epi[ith]*dsig;
  }

  /* compute geopotential */
  for (ith = 0; ith < nth; ith++) {
    z[ith] = (mont[ith] - (SIGC(ith)*(tht0 - thb) + thb)*pi[ith])/GEE;
  }
}

/* integrate vertical profiles on edge grid, yielding edge values */
void vert_jcomp(double thb, double topo, double *xeta, double *xpi,
		double *xmont, double *xz)
{
  int ith;
  double p, pold;

  /* integrate -eta downward to get pi -- values of p on grid edges */
  p = ptop;
  xpi[nth] = pifromp(p);
  for (ith = nth - 1; ith >= 0; ith--) {
    pold = p;
    p = pold + 0.5*GEE*(xeta[ith + 1] + xeta[ith])*dsig;
    xpi[ith] = pifromp(p);
  }

  /* integrate pi upward to get mont */
  xmont[0] = thb*xpi[0] + GEE*topo;
  for (ith = 1; ith <= nth; ith++) {
    xmont[ith] = xmont[ith - 1] +
      0.5*(tht0 - thb)*(xpi[ith] + xpi[ith - 1])*dsig;
  }

  /* compute geopotential */
  for (ith = 0; ith <= nth; ith++) {
    xz[ith] = (xmont[ith] - (thb + ith*dsig*(tht0 - thb))*xpi[ith])/GEE;
  }
}
