/* wallclock -- collect wall times at various points in program */
#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>

int *wallsec, *wallmicro;
int *walllist, wallnum, wallmax;

/* initialization routine to allocate arrays */
void wallinit(int wallmaxin)
{
  wallnum = 0;
  wallmax = wallmaxin;
  if ((wallsec = (int*)malloc((unsigned)(sizeof(int)*wallmax))) ==
      (int*)NULL) {
    fprintf(stderr,"wallclock: can't get space\n");
    exit(1);
  }
  if ((wallmicro = (int*)malloc((unsigned)(sizeof(int)*wallmax))) ==
      (int*)NULL) {
    fprintf(stderr,"wallclock: can't get space\n");
    exit(1);
  }
  if ((walllist = (int*)malloc((unsigned)(sizeof(int)*wallmax))) ==
      (int*)NULL) {
    fprintf(stderr,"wallclock: can't get space\n");
    exit(1);
  }
}

/* routine to be called at desired wall clock monitoring points */
void wallget(int loc)
{
  struct timeval tv;

  /* do get time of day call and exit if it doesn't succeed */
  if (gettimeofday(&tv, NULL) == -1) {
    fprintf(stderr,"wallget: failed\n");
    exit(1);
  }

  /* decode the time of day and store it in the collection arrays */
  wallsec[wallnum] = tv.tv_sec;
  wallmicro[wallnum] = tv.tv_usec;
  walllist[wallnum] = loc;

  /* die if too many collected points for allocated array size */
  if (wallnum >= wallmax - 1) {
    fprintf(stderr,"wallget: too many collected points\n");
    exit(1);
  }
  else wallnum++;
}

/* write the wall clock output to the standard output */
void wallwrite()
{
  int wallbase, i, j, k, nlocs, nsteps;
  double tvalue = 0., oldvalue;

  /* check for number of points selected */
  nlocs = 0;
  for (i = 0; i < wallnum; i++) {
    if (walllist[i] > nlocs) nlocs = walllist[i];
  }
  nlocs++;

  /* check list for consistency */
  if ((wallnum % nlocs) != 0) {
    fprintf(stderr,"wallwrite: nlocs %d doesn't divide into wallnum %d\n",
	    nlocs,wallnum);
    exit(1);
  }

  /* write output header */
  for (j = 0; j < nlocs; j++) {
    fprintf(stdout,"#     loc %4d ",j);
  }
  fprintf(stdout,"\n");

  /* write output */
  nsteps = wallnum/nlocs;
  wallbase = wallsec[0];
  for (i = 0; i < nsteps; i++) {
    for (j = 0; j < nlocs; j++) {

      /* compute the position in the original table */
      k = i*nlocs + j;

      /* another consistency check */
      if (walllist[k] != j) {
	fprintf(stderr,"wallwrite: inconsistent list at i,j = %d,%d\n",i,j);
	exit(1);
      }

      /* compute the wall clock time value in seconds, saving the old value */
      oldvalue = tvalue;
      tvalue = (double)(wallsec[k] - wallbase) + (double)(1.e-6*wallmicro[k]);

      /* for the first entry print the actual time value -- for
       * the others, print the difference between the current
       * and previous values to get the total wall time between
       * calls to wallget
       */
      if (j == 0) fprintf(stdout,"%14.6f ",tvalue);
      else fprintf(stdout,"%14.6f ",tvalue - oldvalue);
    }
    fprintf(stdout,"\n");
  }

  /* reset for next collection */
  wallnum = 0;
}
